import demes
import numpy as np
import unittest
import pytest
import fwdpy11


def check_valid_demography(cls):
    def _valid_fwdpy11_demography(self):
        try:
            _ = fwdpy11.DemographyDebugger(
                [100] * len(self.demog.metadata["initial_sizes"]), self.demog
            )
        except:
            self.fail("unexpected exception")

    cls.test_validity = _valid_fwdpy11_demography
    return cls


@check_valid_demography
class TestNoEvents(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(name="deme", epochs=[dict(start_size=1000, end_time=0)])

        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


class TestBadBurnin(unittest.TestCase):
    def test_burnin_inputs(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(name="deme", epochs=[dict(start_size=1000, end_time=0)])

        self.g = self.b.resolve()
        with pytest.raises(ValueError):
            self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10.0)
        with pytest.raises(ValueError):
            self.demog = fwdpy11.discrete_demography.from_demes(self.g, -1)


@check_valid_demography
class TestLoadGraph(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.g = demes.load("tests/test_demog.yaml")
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


@check_valid_demography
class TestLoadYAML(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.demog = fwdpy11.discrete_demography.from_demes("tests/test_demog.yaml", 10)


@check_valid_demography
class TestTwoEpoch(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(
            name="deme",
            epochs=[
                dict(start_size=1000, end_time=100),
                dict(start_size=2000, end_time=0),
            ],
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_size_change_params(self):
        self.assertTrue(len(self.demog.model.set_deme_sizes) == 1)
        self.assertTrue(
            self.demog.model.set_deme_sizes[0].when
            == self.demog.metadata["burnin_time"] - 1
        )
        self.assertTrue(self.demog.model.set_deme_sizes[0].new_size == 2000)


@check_valid_demography
class TestNonGenerationUnits(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="years", generation_time=25
        )
        self.b.add_deme(
            name="Pop",
            epochs=[
                dict(start_size=10000, end_time=25000),
                dict(start_size=1000, end_size=20000, end_time=0),
            ],
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_conversion_to_generations(self):
        self.assertTrue(
            self.demog.model.set_deme_sizes[0].when
            == self.demog.metadata["burnin_time"] - 1
        )
        self.assertTrue(
            self.demog.metadata["total_simulation_length"]
            == self.demog.metadata["burnin_time"] + 25000 // 25
        )
        self.assertTrue(self.demog.model.set_deme_sizes[0].new_size == 1000)


@check_valid_demography
class TestSelfingShift(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(
            name="Selfer",
            epochs=[
                dict(start_size=1000, end_time=1000),
                dict(start_size=1000, end_time=0, selfing_rate=0.2),
            ],
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_selfing_parameters(self):
        self.assertTrue(self.demog.model.set_selfing_rates[0].when == 0)
        self.assertTrue(self.demog.model.set_selfing_rates[0].S == 0.0)
        self.assertTrue(
            self.demog.model.set_selfing_rates[1].when
            == self.demog.metadata["burnin_time"]
        )
        self.assertTrue(self.demog.model.set_selfing_rates[1].S == 0.2)


@check_valid_demography
class TestSelfing(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(
            name="Selfer",
            epochs=[dict(start_size=1000, end_time=0)],
            defaults=dict(epoch=dict(selfing_rate=0.5)),
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_single_pop_selfing(self):
        self.assertTrue(self.demog.model.set_selfing_rates[0].when == 0)
        self.assertTrue(self.demog.model.set_selfing_rates[0].S == 0.5)


@check_valid_demography
class TestSplit(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_size_changes(self):
        self.assertTrue(len(self.demog.model.set_deme_sizes) == 3)
        self.assertTrue(self.demog.model.set_deme_sizes[0].deme == 1)
        self.assertTrue(self.demog.model.set_deme_sizes[0].new_size == 100)
        self.assertTrue(
            self.demog.model.set_deme_sizes[0].when
            == self.demog.metadata["burnin_time"] - 1
        )
        self.assertTrue(self.demog.model.set_deme_sizes[1].deme == 2)
        self.assertTrue(self.demog.model.set_deme_sizes[1].new_size == 100)
        self.assertTrue(
            self.demog.model.set_deme_sizes[1].when
            == self.demog.metadata["burnin_time"] - 1
        )
        self.assertTrue(self.demog.model.set_deme_sizes[2].deme == 0)
        self.assertTrue(self.demog.model.set_deme_sizes[2].new_size == 0)
        self.assertTrue(
            self.demog.model.set_deme_sizes[2].when
            == self.demog.metadata["burnin_time"]
        )


@check_valid_demography
class TestSplitMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_migration(source="Deme1", dest="Deme2", rate=0.01)
        self.b.add_migration(source="Deme2", dest="Deme1", rate=0.02)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


@check_valid_demography
class TestSplitSymmetricMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_migration(demes=["Deme1", "Deme2"], rate=0.01)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


@check_valid_demography
class TestSplitThreeWay(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme3", epochs=[dict(start_size=200, end_time=0)], ancestors=["Ancestor"]
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


@check_valid_demography
class TestSplitThreeWayMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme3", epochs=[dict(start_size=200, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_migration(demes=["Deme1", "Deme2", "Deme3"], rate=0.1)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


@check_valid_demography
class TestBranch(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test branch", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[dict(start_size=1000, end_time=0)])
        self.b.add_deme(
            "Deme1",
            epochs=[dict(start_size=100, end_time=0)],
            ancestors=["Ancestor"],
            start_time=100,
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


@check_valid_demography
class TestBranchMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test branch", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[dict(start_size=1000, end_time=0)])
        self.b.add_deme(
            "Deme1",
            epochs=[dict(start_size=100, end_time=0)],
            ancestors=["Ancestor"],
            start_time=100,
        )
        self.b.add_migration(source="Ancestor", dest="Deme1", rate=0.01)
        self.b.add_migration(source="Deme1", dest="Ancestor", rate=0.01)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


@check_valid_demography
class TestMultipleBranches(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test branch", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[dict(start_size=1000, end_time=0)])
        self.b.add_deme(
            "Deme1",
            epochs=[dict(start_size=100, end_time=0)],
            ancestors=["Ancestor"],
            start_time=100,
        )
        self.b.add_deme(
            "Deme2",
            epochs=[dict(start_size=200, end_time=0)],
            ancestors=["Ancestor"],
            start_time=50,
        )
        self.b.add_deme(
            "Deme3",
            epochs=[dict(start_size=300, end_time=0)],
            ancestors=["Deme1"],
            start_time=20,
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


@check_valid_demography
class TestSplitsBranches(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test", time_units="generations")
        self.b.add_deme(name="A", epochs=[dict(start_size=1000, end_time=100)])
        self.b.add_deme(
            name="B",
            epochs=[dict(start_size=1000, end_time=0)],
            ancestors=["A"],
            start_time=200,
        )
        self.b.add_deme(
            name="C", epochs=[dict(start_size=1000, end_time=0)], ancestors=["A"]
        )
        self.b.add_deme(
            name="D", epochs=[dict(start_size=1000, end_time=0)], ancestors=["A"]
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)


@check_valid_demography
class TestIslandModel(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="island", time_units="generations")
        self.b.add_deme(name="Island1", epochs=[dict(start_size=100, end_time=0)])
        self.b.add_deme(name="Island2", epochs=[dict(start_size=200, end_time=0)])
        self.b.add_migration(source="Island1", dest="Island2", rate=0.01)
        self.b.add_migration(source="Island2", dest="Island1", rate=0.02)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_demog_attributes(self):
        self.assertTrue(
            self.demog.metadata["burnin_time"]
            == sum(self.demog.metadata["initial_sizes"].values()) * 10
        )
        self.assertTrue(len(self.demog.model.set_migration_rates) == 0)


@check_valid_demography
class TestIslandModelRateChange(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="island", time_units="generations")
        self.b.add_deme(name="Island1", epochs=[dict(start_size=100, end_time=0)])
        self.b.add_deme(name="Island2", epochs=[dict(start_size=200, end_time=0)])
        self.b.add_migration(source="Island1", dest="Island2", rate=0.01, end_time=500)
        self.b.add_migration(source="Island2", dest="Island1", rate=0.02)
        self.b.add_migration(
            source="Island1", dest="Island2", rate=0.05, start_time=500
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_burnin_time(self):
        self.assertTrue(
            self.demog.metadata["burnin_time"]
            == sum(self.demog.metadata["initial_sizes"].values()) * 10
        )

    def test_num_mig_rate_changes(self):
        self.assertTrue(len(self.demog.model.set_migration_rates) == 1)

    def test_total_sim_length(self):
        self.assertTrue(
            self.demog.metadata["total_simulation_length"]
            == self.demog.metadata["burnin_time"] + 500
        )


@check_valid_demography
class TestTwoPopMerger(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="split then merger", time_units="generations"
        )
        self.b.add_deme(name="Ancestral", epochs=[dict(start_size=1000, end_time=1000)])
        self.b.add_deme(
            name="Parent1",
            epochs=[dict(start_size=500, end_time=500)],
            ancestors=["Ancestral"],
        )
        self.b.add_deme(
            name="Parent2",
            epochs=[dict(start_size=500, end_time=500)],
            ancestors=["Ancestral"],
        )
        self.b.add_deme(
            name="Child",
            epochs=[dict(start_size=1000, end_time=0)],
            ancestors=["Parent1", "Parent2"],
            proportions=[0.5, 0.5],
            start_time=500,
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_total_sim_length(self):
        self.assertTrue(
            self.demog.metadata["total_simulation_length"]
            == self.demog.metadata["burnin_time"] + 1000
        )

    def test_num_size_changes(self):
        self.assertTrue(len(self.demog.model.set_deme_sizes) == 6)


@check_valid_demography
class TestFourWayMerger(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="split then merger", time_units="generations"
        )
        self.b.add_deme(name="Ancestral", epochs=[dict(start_size=1000, end_time=1000)])
        self.b.add_deme(
            name="A",
            epochs=[dict(start_size=500, end_time=700)],
            ancestors=["Ancestral"],
        )
        self.b.add_deme(
            name="B",
            epochs=[dict(start_size=500, end_time=500)],
            ancestors=["Ancestral"],
        )
        self.b.add_deme(
            name="Parent1", epochs=[dict(start_size=200, end_time=100)], ancestors=["A"]
        )
        self.b.add_deme(
            name="Parent2", epochs=[dict(start_size=300, end_time=100)], ancestors=["A"]
        )
        self.b.add_deme(
            name="Parent3", epochs=[dict(start_size=100, end_time=100)], ancestors=["B"]
        )
        self.b.add_deme(
            name="Parent4", epochs=[dict(start_size=400, end_time=100)], ancestors=["B"]
        )
        self.b.add_deme(
            name="Child",
            epochs=[dict(start_size=1000, end_time=0)],
            ancestors=["Parent1", "Parent2", "Parent3", "Parent4"],
            proportions=[0.2, 0.3, 0.1, 0.4],
            start_time=100,
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_total_sim_length(self):
        self.assertTrue(
            self.demog.metadata["total_simulation_length"]
            == self.demog.metadata["burnin_time"] + 1000
        )

    def test_num_size_changes(self):
        self.assertTrue(len(self.demog.model.set_deme_sizes) == 14)


@check_valid_demography
class TestPulseMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="pulse", time_units="generations")
        self.b.add_deme(name="deme1", epochs=[dict(start_size=100, end_time=0)])
        self.b.add_deme(name="deme2", epochs=[dict(start_size=100, end_time=0)])
        self.b.add_pulse(source="deme1", dest="deme2", time=100, proportion=0.2)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10)

    def test_total_sim_length(self):
        self.assertTrue(
            self.demog.metadata["total_simulation_length"]
            == self.demog.metadata["burnin_time"] + 100
        )

    def test_pulse_migration_matrix(self):
        self.assertTrue(len(self.demog.model.set_migration_rates) == 2)
        self.assertTrue(
            self.demog.model.set_migration_rates[0].when
            == self.demog.metadata["burnin_time"] - 1
        )
        self.assertTrue(
            self.demog.model.set_migration_rates[1].when
            == self.demog.metadata["burnin_time"]
        )
        self.assertTrue(self.demog.model.set_migration_rates[0].deme == 1)
        self.assertTrue(self.demog.model.set_migration_rates[1].deme == 1)
        self.assertTrue(
            np.all(self.demog.model.set_migration_rates[0].migrates == [0.2, 0.8])
        )
        self.assertTrue(
            np.all(self.demog.model.set_migration_rates[1].migrates == [0, 1])
        )



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import typing

import attr
from fwdpy11.class_decorators import (attr_add_asblack, attr_class_pickle,
                                      attr_class_to_from_dict,
                                      attr_class_to_from_dict_no_recurse)


@attr_add_asblack
@attr_class_pickle
@attr_class_to_from_dict
@attr.s(
    kw_only=True, frozen=True, auto_attribs=True, repr_ns="fwdpy11.demographic_models"
)
class DemographicModelCitation(object):
    """
    Citation information for a demographic model

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param DOI: The Digital Object Identifier
    :param full_citation: Something string-like giving the full citation.
    :param metadata: Any additional information that may be needed.


    .. versionadded:: 0.8.0
    """

    DOI: object
    full_citation: object
    metadata: object


@attr_add_asblack
@attr_class_pickle
@attr_class_to_from_dict_no_recurse
@attr.s(
    kw_only=True, frozen=True, auto_attribs=True, repr_ns="fwdpy11.demographic_models"
)
class DemographicModelDetails(object):
    """
    Stores rich information about a demographic model.
    Instances of this class get returned by functions
    generating pre-calculated models.

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param model: The demographic model parameters
    :type model: object
    :param name: The name of the model
    :type name: str
    :param source: The source of the model
    :type source: dict
    :param parameters: The parameters used to generate ``model``
    :type parameters: object
    :param citation: Citation information
    :type citation: dict or fwdpy11.demographic_models.DemographicModelCitation or None
    :param metadata: Optional field for additional info
    :type metadata: object

    .. versionadded:: 0.8.0
    """

    model: object
    name: str
    source: typing.Dict
    parameters: object
    citation: typing.Optional[typing.Union[DemographicModelCitation, typing.Dict]]
    metadata: typing.Optional[object] = None



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import typing

import attr
import tskit

from .._fwdpy11 import Mutation, MutationVector
from ._flags import (INDIVIDUAL_IS_ALIVE, INDIVIDUAL_IS_FIRST_GENERATION,
                     INDIVIDUAL_IS_PRESERVED)


@attr.s(auto_attribs=True, frozen=True, kw_only=True, eq=True)
class DiploidMetadata(object):
    """
    Instances of this type are returned by
    :func:`fwdpy11.tskit_tools.decode_individual_metadata`.

    The following ``kwargs`` are used to construct instances of
    this type. Member attributes have the same names as these
    ``kwargs```:

    :param g: Genetic value
    :type g: float
    :param e: "Random/environmental" value
    :type e: float
    :param w: Fitness
    :type w: float
    :param sex: Sex
    :type sex: int
    :param deme: Deme
    :type deme: int
    :param label: Index in the population at time of living/sampling
    :type label: int
    :param alive: Individual is alive
    :type alive: bool
    :param preserved: Individual is preserved
    :type preserved: bool
    :param first_generation: Individual is first generation
    :type first_generation: bool
    :param parents: ``label`` of parents
    :type parents: list
    :param geography: Location in 3d space
    :type geography: list
    :param nodes: Nodes on tree sequence
    :type nodes: list

    .. versionadded:: 0.12.0
    """

    g: float
    e: float
    w: float
    sex: int
    deme: int
    label: int
    alive: bool
    preserved: bool
    first_generation: bool
    parents: typing.List[int]
    geography: typing.List[float]
    nodes: typing.List[int]


def decode_individual_metadata(tc: tskit.TableCollection):
    """
    Decodes a :class:`tskit.IndividualTable`.

    :param tc: A table collection
    :type tc: :class:`tskit.TableCollection`

    :returns: individual metadata
    :rtype: List of :class:`fwdpy11.tskit_tools.DiploidMetadata`


    .. versionadded:: 0.12.0
    """
    rv = []
    for i in tc.individuals:
        alive = i.flags & INDIVIDUAL_IS_ALIVE
        preserved = i.flags & INDIVIDUAL_IS_PRESERVED
        first_generation = i.flags & INDIVIDUAL_IS_FIRST_GENERATION
        rv.append(
            DiploidMetadata(
                **i.metadata,
                alive=alive,
                preserved=preserved,
                first_generation=first_generation
            )
        )

    return rv


def decode_mutation_metadata(tc: tskit.TableCollection) -> MutationVector:
    """
    Decodes a :class:`tskit.MutationTable`.

    :param tc: A table collection
    :type tc: :class:`tskit.TableCollection`

    :returns: Mutations
    :rtype: :class:`fwdpy11.MutationVector`

    .. versionadded:: 0.12.0
    """
    mutations = MutationVector()
    for m in tc.mutations:
        md = m.metadata
        if "esizes" not in md:
            mutations.append(
                Mutation(
                    pos=tc.sites.position[m.site],
                    s=md["s"],
                    h=md["h"],
                    g=md["origin"],
                    label=md["label"],
                )
            )
        else:
            mutations.append(
                Mutation(
                    pos=tc.sites.position[m.site],
                    s=md["s"],
                    h=md["h"],
                    g=md["origin"],
                    esizes=md["esizes"],
                    heffects=md["heffects"],
                    label=md["label"],
                )
            )
    return mutations



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import pytest

import fwdpy11


@pytest.fixture(scope="function")
def rng(request):
    try:
        seed = request.param.get("seed", None)
        return fwdpy11.GSLrng(seed)
    except AttributeError:  # NOQA
        return fwdpy11.GSLrng(42)


@pytest.fixture(scope="function")
def pop(request):
    popsize = request.param["N"]
    genome_length = request.param["genome_length"]
    return fwdpy11.DiploidPopulation(popsize, genome_length)


@pytest.fixture(scope="function")
def mslike_pop(request):
    try:
        N = request.param.get("N", None)
        return fwdpy11.DiploidPopulation(N, 1.0)
    except AttributeError:  # NOQA
        return fwdpy11.DiploidPopulation(1000, 1.0)



from .data_matrix_from_tables import (data_matrix_from_tables,  # NOQA
                                      make_data_matrix)
from .import_demes import demography_from_demes  # NOQa
from .simplify_tables import simplify, simplify_tables  # NOQA



#
# Copyright (C) 2017-2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

"""
The concepts tested here are redundant with other test files.
The goal is to demonstrate specific examples that are hopefully
digestible and can serve as examples.
"""

import typing
import unittest

import attr
import numpy as np

import fwdpy11


@attr.s(auto_attribs=True)
class NumberOfParentsAtTime(object):
    generation: int
    deme: int
    size: int


@attr.s(auto_attribs=True)
class DemeSizeAtTime(object):
    generation: int
    deme: int
    size: int


@attr.s(auto_attribs=True)
class DemeSizeTracker(object):
    parent_numbers: typing.List[NumberOfParentsAtTime]
    individual_numbers: typing.List[DemeSizeAtTime]

    def __call__(
        self, pop: fwdpy11.DiploidPopulation, sampler: fwdpy11.SampleRecorder
    ) -> None:
        for i in pop.diploid_metadata:
            for n in i.nodes:
                assert (
                    pop.tables.nodes[n].time == pop.generation
                ), f"{pop.tables.nodes[n].time}, {pop.generation}"
        deme_sizes = pop.deme_sizes()
        for i, j in zip(deme_sizes[0], deme_sizes[1]):
            self.parent_numbers.append(NumberOfParentsAtTime(pop.generation, i, j))

        number_of_individuals = np.unique(
            [i.deme for i in pop.diploid_metadata[: pop.N]], return_counts=True
        )
        for i, j in zip(number_of_individuals[0], number_of_individuals[1]):
            self.individual_numbers.append(DemeSizeAtTime(pop.generation, i, j))


@attr.s(auto_attribs=True)
class ParentData(object):
    offspring_generation: int
    offspring_deme: int
    parent1: int
    parent2: int


@attr.s(auto_attribs=True)
class TrackParents(object):
    parents: typing.List[ParentData]

    def __call__(
        self, pop: fwdpy11.DiploidPopulation, sampler: fwdpy11.SampleRecorder
    ) -> None:
        for i in pop.diploid_metadata:
            self.parents.append(
                ParentData(pop.generation, i.deme, i.parents[0], i.parents[1])
            )


def branch(when: int, fraction: float) -> fwdpy11.DiscreteDemography:
    """
    A deme gives rise to a new daugther deme.
    The original deme continues to exist.
    """
    mass_migrations = [
        fwdpy11.copy_individuals(when=when, source=0, destination=1, fraction=fraction)
    ]
    return fwdpy11.DiscreteDemography(mass_migrations=mass_migrations)


def split(when: int, fraction: float) -> fwdpy11.DiscreteDemography:
    """
    Parental deme 0 gives rise to daughter demes 1 and 2.
    Deme 0 ceases to exist.

    Daughter demes are created by copying parents.
    """
    mass_migrations = [
        fwdpy11.copy_individuals(when=when, source=0, destination=1, fraction=fraction),
        fwdpy11.copy_individuals(
            when=when, source=0, destination=2, fraction=1.0 - fraction
        ),
    ]
    set_deme_sizes = [fwdpy11.SetDemeSize(when=when, deme=0, new_size=0)]
    return fwdpy11.DiscreteDemography(
        mass_migrations=mass_migrations, set_deme_sizes=set_deme_sizes
    )


def split_via_moves(when: int, fraction: float) -> fwdpy11.DiscreteDemography:
    """
    Parental deme 0 gives rise to daughter demes 1 and 2.
    Deme 0 ceases to exist.

    Daughter demes are created by moving parents.
    """
    mass_migrations = [
        fwdpy11.move_individuals(when=when, source=0, destination=1, fraction=fraction),
        fwdpy11.move_individuals(
            when=when, source=0, destination=2, fraction=1.0 - fraction
        ),
    ]
    return fwdpy11.DiscreteDemography(mass_migrations=mass_migrations)


def merge(
    when: int, p0: float, p1: float, keep_ancestor_demes: bool
) -> fwdpy11.DiscreteDemography:
    """
    Demes 0 and 1 merge to form deme 2, which is formed by fractions
    p0 and p1 of demes 0 and 1, respectively.

    Whether or not to "kill off" the ancestral demes is an input choice
    """
    mass_migrations = [
        fwdpy11.copy_individuals(when=when, source=0, destination=2, fraction=p0),
        fwdpy11.copy_individuals(when=when, source=1, destination=2, fraction=p1),
    ]
    if keep_ancestor_demes is False:
        set_deme_sizes = [
            fwdpy11.SetDemeSize(when=when, deme=0, new_size=0),
            fwdpy11.SetDemeSize(when=when, deme=1, new_size=0),
        ]
    else:
        set_deme_sizes = []
    return fwdpy11.DiscreteDemography(
        mass_migrations=mass_migrations, set_deme_sizes=set_deme_sizes
    )


def pulse_migration(
    when: int, sizes_when: typing.List[int], ancestry_fraction: float
) -> fwdpy11.DiscreteDemography:
    """
    At time ``when``, ``ancestry_fraction`` of deme 1
    comes from deme 0, meaning that the offspring born in generation
    ``when + 1`` have a parent from 0 with probability ``ancestry_fraction``
    and a parent from 1 with probability ``1. - ancestry_fraction``.

    The size of deme 1 remains unchanged.

    To get this right, we need to know the deme sizes at the time of
    the pulse migration.  These sizes let us calculate how many individuals
    to copy from i to j to be potential parents of generation ``when + 1``.
    """
    parents_needed_from_deme_0 = np.rint(
        ancestry_fraction * sizes_when[1] / (1.0 - ancestry_fraction)
    ).astype(int)
    copy_fraction = parents_needed_from_deme_0 / sizes_when[0]

    mass_migrations = [
        fwdpy11.copy_individuals(
            when=when, source=0, destination=1, fraction=copy_fraction
        )
    ]
    set_deme_sizes = [fwdpy11.SetDemeSize(when=when, deme=1, new_size=sizes_when[1])]
    return fwdpy11.DiscreteDemography(
        mass_migrations=mass_migrations, set_deme_sizes=set_deme_sizes
    )


def pulse_migration_via_setting_migration_rates(
    when: int, source: int, destination: int, fraction: float
) -> fwdpy11.DiscreteDemography:
    """
    An arguably simpler implementation of pulse migration.
    The migration matrix is interpreted as the ancestry proportion
    from each input deme each generation.

    At generation when, we can change the migration inputs into deme
    ``destination`` and then reset them at generation when + 1.

    The outcome is that the offspring born at time when + 1 are the
    result of the pulse migration and offspring born at time when + 2
    have the reset ancestry proportions.

    Below, we test this function by setting source = 0, destination = 1,
    fraction = 1.0, meaning that deme 1 is replaced by deme 0, meaning
    that all parents of offspring born at time when + 1 must be
    from deme 0.
    """
    migmatrix = np.eye(2)
    m = np.zeros(2)
    m[source] = fraction
    m[destination] = 1.0 - fraction
    set_migration_rates = [
        fwdpy11.SetMigrationRates(when=when, deme=destination, migrates=m)
    ]
    set_migration_rates.append(
        fwdpy11.SetMigrationRates(
            when=when + 1, deme=destination, migrates=migmatrix[destination, :]
        )
    )
    return fwdpy11.DiscreteDemography(
        migmatrix=migmatrix, set_migration_rates=set_migration_rates
    )


def build_params(
    demography: fwdpy11.DiscreteDemography, simlen: int
) -> fwdpy11.ModelParams:
    pdict = {
        "nregions": [],
        "sregions": [],
        "recregions": [],
        "rates": [0.0, 0.0, 0.0],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "demography": demography,
        "simlen": simlen,
    }
    return fwdpy11.ModelParams(**pdict)


def run_model(
    params: fwdpy11.ModelParams, popsizes: typing.Union[int, typing.List], seed: int
) -> typing.Tuple[
    fwdpy11.DiploidPopulation,
    typing.List[NumberOfParentsAtTime],
    typing.List[DemeSizeAtTime],
]:
    # 1.0 is the genome length for the tree sequence recording
    pop = fwdpy11.DiploidPopulation(popsizes, 1.0)
    rng = fwdpy11.GSLrng(seed)
    tracker = DemeSizeTracker([], [])
    fwdpy11.evolvets(rng, pop, params, 10, tracker)  # simplify every 10 generations
    return pop, tracker.parent_numbers, tracker.individual_numbers


def run_model_with_parent_tracking(
    params: fwdpy11.ModelParams, popsizes: typing.Union[int, typing.List], seed: int
) -> typing.Tuple[fwdpy11.DiploidPopulation, typing.List[ParentData]]:
    # 1.0 is the genome length for the tree sequence recording
    pop = fwdpy11.DiploidPopulation(popsizes, 1.0)
    rng = fwdpy11.GSLrng(seed)
    tracker = TrackParents([])
    fwdpy11.evolvets(rng, pop, params, 10, tracker)  # simplify every 10 generations
    return pop, tracker.parents


class TestBranch(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.when = 10
        self.fraction = 0.25
        self.popsizes = 100
        demography = branch(self.when, self.fraction)
        self.params = build_params(demography, 50)
        self.pop, self.num_parents, self.num_individuals = run_model(
            self.params, self.popsizes, 666
        )

    def test_final_sizes(self):
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], self.popsizes)
        self.assertEqual(deme_sizes[1], np.rint(self.fraction * self.popsizes))

    def test_parent_sizes_over_time(self):
        self.assertEqual(
            len([i for i in self.num_parents if i.deme == 0]), self.params.simlen
        )
        # Deme 1 first appeared in generation 10, so there were 9 generations
        # w/o that deme around.
        self.assertEqual(
            len([i for i in self.num_parents if i.deme == 1]),
            self.params.simlen - self.when + 1,
        )

    def test_deme_sizes_over_time(self):
        self.assertEqual(
            len([i for i in self.num_individuals if i.deme == 0]), self.params.simlen
        )
        self.assertEqual(
            len([i for i in self.num_individuals if i.deme == 1]),
            self.params.simlen - self.when,
        )


class TestSplit(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.when = 10
        self.fraction = 0.25
        self.popsizes = 100
        demography = split(self.when, self.fraction)
        self.params = build_params(demography, 50)
        self.pop, self.num_parents, self.num_individuals = run_model(
            self.params, self.popsizes, 666
        )

    def test_final_sizes(self):
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertTrue(0 not in deme_sizes)
        self.assertEqual(deme_sizes[1], np.rint(self.fraction * self.popsizes))
        self.assertEqual(deme_sizes[2], np.rint((1.0 - self.fraction) * self.popsizes))

    def test_parent_sizes_over_time(self):
        self.assertEqual(
            max([i.generation for i in self.num_parents if i.deme == 0]), self.when
        )
        self.assertEqual(
            min([i.generation for i in self.num_parents if i.deme == 1]), self.when
        )
        self.assertEqual(
            min([i.generation for i in self.num_parents if i.deme == 2]), self.when
        )

    def test_deme_sizes_over_time(self):
        self.assertEqual(
            max([i.generation for i in self.num_individuals if i.deme == 0]), self.when
        )
        self.assertEqual(
            min([i.generation for i in self.num_individuals if i.deme == 1]),
            self.when + 1,
        )
        self.assertEqual(
            min([i.generation for i in self.num_individuals if i.deme == 2]),
            self.when + 1,
        )


class TestSplitViaMoves(unittest.TestCase):
    # This model differs from the above in that
    # no parents will exist in deme 0 at time self.when - 1
    # because they were all moved out to create the other demes.
    @classmethod
    def setUpClass(self):
        self.when = 10
        self.fraction = 0.25
        self.popsizes = 100
        demography = split_via_moves(self.when, self.fraction)
        self.params = build_params(demography, 50)
        self.pop, self.num_parents, self.num_individuals = run_model(
            self.params, self.popsizes, 666
        )

    def test_final_sizes(self):
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertTrue(0 not in deme_sizes)
        self.assertEqual(deme_sizes[1], np.rint(self.fraction * self.popsizes))
        self.assertEqual(deme_sizes[2], np.rint((1.0 - self.fraction) * self.popsizes))

    def test_parent_sizes_over_time(self):
        self.assertEqual(
            max([i.generation for i in self.num_parents if i.deme == 0]), self.when - 1
        )
        self.assertEqual(
            min([i.generation for i in self.num_parents if i.deme == 1]), self.when
        )
        self.assertEqual(
            min([i.generation for i in self.num_parents if i.deme == 2]), self.when
        )

    def test_deme_sizes_over_time(self):
        self.assertEqual(
            max([i.generation for i in self.num_individuals if i.deme == 0]),
            self.when - 1,
        )
        self.assertEqual(
            min([i.generation for i in self.num_individuals if i.deme == 1]), self.when
        )
        self.assertEqual(
            min([i.generation for i in self.num_individuals if i.deme == 2]), self.when
        )


class TestMergeKeepAncestralDemes(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.when = 10
        self.p0 = 0.33
        self.p1 = 0.77
        self.popsizes = [100, 100]
        demography = merge(self.when, self.p0, self.p1, True)
        self.params = build_params(demography, 50)
        self.pop, self.num_parents, self.num_individuals = run_model(
            self.params, self.popsizes, 666
        )

    def test_final_sizes(self):
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], self.popsizes[0])
        self.assertEqual(deme_sizes[1], self.popsizes[1])
        self.assertEqual(
            deme_sizes[2],
            np.rint(self.p0 * self.popsizes[0]) + np.rint(self.p1 * self.popsizes[1]),
        )

    def test_parent_sizes_over_time(self):
        self.assertEqual(
            len([i for i in self.num_parents if i.deme == 0]), self.params.simlen
        )
        self.assertEqual(
            len([i for i in self.num_parents if i.deme == 1]), self.params.simlen
        )

        min_deme_2 = min([i.generation for i in self.num_parents if i.deme == 2])

        self.assertEqual(min_deme_2, self.when)

        # Deme 2 parents first appeared in generation 10, so there were 9 generations
        # w/o any
        self.assertEqual(
            len([i for i in self.num_parents if i.deme == 2]),
            self.params.simlen - self.when + 1,
        )

    def test_deme_sizes_over_time(self):
        self.assertEqual(
            len([i for i in self.num_individuals if i.deme == 0]), self.params.simlen
        )
        self.assertEqual(
            len([i for i in self.num_individuals if i.deme == 1]), self.params.simlen
        )
        min_deme_2 = min([i.generation for i in self.num_individuals if i.deme == 2])

        self.assertEqual(min_deme_2, self.when + 1)

        self.assertEqual(
            len([i for i in self.num_individuals if i.deme == 2]),
            self.params.simlen - self.when,
        )


class TestMergeKillAncestralDemes(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.when = 10
        self.p0 = 0.33
        self.p1 = 0.77
        self.popsizes = [100, 100]
        demography = merge(self.when, self.p0, self.p1, False)
        self.params = build_params(demography, 50)
        self.pop, self.num_parents, self.num_individuals = run_model(
            self.params, self.popsizes, 666
        )

    def test_final_sizes(self):
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertTrue(0 not in deme_sizes)
        self.assertTrue(1 not in deme_sizes)
        self.assertEqual(
            deme_sizes[2],
            np.rint(self.p0 * self.popsizes[0]) + np.rint(self.p1 * self.popsizes[1]),
        )

    def test_parent_sizes_over_time(self):
        self.assertEqual(len([i for i in self.num_parents if i.deme == 0]), self.when)
        self.assertEqual(len([i for i in self.num_parents if i.deme == 1]), self.when)

        min_deme_2 = min([i.generation for i in self.num_parents if i.deme == 2])

        self.assertEqual(min_deme_2, self.when)

        # Deme 2 parents first appeared in generation 10, so there were 9 generations
        # w/o any
        self.assertEqual(
            len([i for i in self.num_parents if i.deme == 2]),
            self.params.simlen - self.when + 1,
        )

    def test_deme_sizes_over_time(self):
        self.assertEqual(
            len([i for i in self.num_individuals if i.deme == 0]), self.when
        )
        self.assertEqual(
            len([i for i in self.num_individuals if i.deme == 1]), self.when
        )

        min_deme_2 = min([i.generation for i in self.num_individuals if i.deme == 2])

        self.assertEqual(min_deme_2, self.when + 1)

        # Deme 2 parents first appeared in generation 10, so there were 9 generations
        # w/o any
        self.assertEqual(
            len([i for i in self.num_individuals if i.deme == 2]),
            self.params.simlen - self.when,
        )


class TestPulseMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.when = 10
        self.ancestry_fraction = 0.23
        self.popsizes = [200, 100]
        demography = pulse_migration(self.when, self.popsizes, self.ancestry_fraction)
        self.params = build_params(demography, 50)
        self.pop, self.num_parents, self.num_individuals = run_model(
            self.params, self.popsizes, 666
        )

    def test_final_sizes(self):
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        for key, value in deme_sizes.items():
            self.assertEqual(value, self.popsizes[key])

    def test_parent_sizes_over_time(self):
        self.assertTrue(
            all([i.size == self.popsizes[0] for i in self.num_parents if i.deme == 0])
        )

        self.assertTrue(
            all(
                [
                    i.size == self.popsizes[1]
                    for i in self.num_parents
                    if i.deme == 1 and i.generation != self.when
                ]
            )
        )

    def test_deme_sizes_over_time(self):
        self.assertTrue(
            all(
                [
                    i.size == self.popsizes[0]
                    for i in self.num_individuals
                    if i.deme == 0
                ]
            )
        )

        self.assertTrue(
            all(
                [
                    i.size == self.popsizes[1]
                    for i in self.num_individuals
                    if i.deme == 1
                ]
            )
        )


class TestPulseMigrationViaSettingMigrationRates(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.when = 10
        self.ancestry_fraction = 1.0
        self.popsizes = [200, 100]
        demography = pulse_migration_via_setting_migration_rates(
            self.when, 0, 1, self.ancestry_fraction
        )
        self.params = build_params(demography, 50)
        self.pop, self.parents = run_model_with_parent_tracking(
            self.params, self.popsizes, 666
        )

    def test_final_sizes(self):
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        for key, value in deme_sizes.items():
            self.assertEqual(value, self.popsizes[key])

    def test_parents_generation_when_plus_one(self):
        plist = [i for i in self.parents if i.offspring_generation == self.when + 1]
        self.assertEqual(len(plist), self.pop.N)
        for p in plist:
            self.assertTrue(p.parent1 < self.popsizes[0])
            self.assertTrue(p.parent2 < self.popsizes[0])

    def test_parents_all_other_generations(self):
        plist = [i for i in self.parents if i.offspring_generation != self.when + 1]
        self.assertEqual(len(plist), self.pop.N * (self.params.simlen - 1))
        for p in plist:
            if p.offspring_deme == 0:
                self.assertTrue(p.parent1 < self.popsizes[0])
                self.assertTrue(p.parent2 < self.popsizes[0])
            else:
                self.assertTrue(p.parent1 >= self.popsizes[0])
                self.assertTrue(p.parent2 >= self.popsizes[0])


if __name__ == "__main__":
    unittest.main()



from typing import IO, Dict, Iterable, Iterator, List, Optional, Tuple, Union

import fwdpy11._types
import fwdpy11.tskit_tools._dump_tables_to_tskit
import numpy as np
import tskit

from .._fwdpy11 import DiploidGenotype, DiploidMetadata, ll_DiploidPopulation
from .population_mixin import PopulationMixin
from .table_collection import TableCollection


class DiploidPopulation(ll_DiploidPopulation, PopulationMixin):
    """
    A population of diploid individuals.

    To initialize a population of size N and genome length L:

    fwdpy11.DiploidPopulation(N, L)

    To initialize a population with two demes of initial
    sizes N1 and N2 and genome length L:

    fwdpy11.DiploidPopulation([N1, N2], L)
    """

    def __init__(
        self,
        N: Union[int, List[int]],
        length: float,
        *,
        ll_pop: Optional[ll_DiploidPopulation] = None
    ):
        if ll_pop is None:
            super(DiploidPopulation, self).__init__(N, length)
        else:
            super(DiploidPopulation, self).__init__(ll_pop)

        self._pytables = TableCollection(self._tables)

    @classmethod
    def create_from_tskit(cls, ts: tskit.TreeSequence):
        """
        Create a new object from an tskit.TreeSequence

        :param ts: A tree sequence from tskit
        :type ts: tskit.TreeSequence

        :return: A population object with an initialized :class:`fwdpy11.TableCollection`
        :rtype: :class:`fwdpy11.DiploidPopulation`

        .. versionadded:: 0.2.0

        .. note::

            In general, initializing a population using
            the output from a coalescent simulation is
            a tricky business.  There are issues of
            parameter scaling and the appropriateness
            of the coalescent model itself. A key issue
            is that your input tree sequence must have
            node times in the correct time units! (Generations,
            for example.)

        """
        ll = ll_DiploidPopulation._create_from_tskit(ts)
        return cls(0, 0.0, ll_pop=ll)

    @classmethod
    def load_from_file(cls, filename: str):
        """
        Load population from the output of
        :func:`fwdpy11.DiploidPopulation.dump_to_file`.

        :param filename: A file name
        :type filename: str
        """
        ll = ll_DiploidPopulation._load_from_file(filename)
        return cls(0, 0.0, ll_pop=ll)

    @classmethod
    def load_from_pickle_file(cls, filename: IO):
        """
        Read in a pickled population from a file.
        The file muse have been generated by
        a call to :func:`fwdpy11.DiploidPopulation.pickle_to_file`.

        :param f: A handle to a file opened in 'rb' mode.

        .. versionadded: 0.3.0

        """
        ll = ll_DiploidPopulation._load_from_pickle_file(filename)
        return cls(0, 0.0, ll_pop=ll)

    @property
    def alive_nodes(self) -> np.ndarray:
        """
        List of alive nodes corresponding to individuals.

        .. versionadded:: 0.5.3
        """
        md = np.array(self.diploid_metadata, copy=False)
        return md["nodes"].flatten()

    @property
    def ancient_sample_nodes(self) -> np.ndarray:
        """
        Return an array of nodes associated with preserved/ancient samples.

        Alias for :attr:`fwdpy11.DiploidPopulation.preserved_nodes`.

        .. versionadded:: 0.13.0
        """
        return self.preserved_nodes

    @property
    def ancient_sample_metadata(self) -> Iterable[DiploidMetadata]:
        """Supports buffer protocol"""
        return self._ancient_sample_metadata

    @property
    def diploids(self) -> Iterable[DiploidGenotype]:
        """Supports buffer protocol"""
        return self._diploids

    @property
    def diploid_metadata(self) -> Iterable[DiploidMetadata]:
        """Supports buffer protocol"""
        return self._diploid_metadata

    @property
    def preserved_nodes(self) -> np.ndarray:
        """
        Return an array of nodes associated with preserved/ancient samples.

        :rtype: :class:`np.ndarray`

        .. versionadded:: 0.13.0

        """
        return np.array(self.ancient_sample_metadata, copy=False)["nodes"].flatten()

    @property
    def tables(self) -> TableCollection:
        """Access the :class:`fwdpy11.TableCollection`"""
        return self._pytables

    def _get_times(self):
        amd = np.array(self.ancient_sample_metadata, copy=False)
        nodes = np.array(self.tables.nodes, copy=False)
        times = nodes["time"][amd["nodes"][:, 0]]
        utimes = np.unique(times)
        return times, utimes

    def deme_sizes(self, as_dict=False) -> Union[np.ndarray, Dict]:
        """
        Return the number of individuals in each deme.

        :param as_dict: If True, return results as a dict.
                        If False, numpy arrays are returned

        The default behavior of this function is equivalent to:

        .. code-block:: python

            md = numpy.array(self.diploid_metadata, copy=False)
            return numpy.unique(md['deme'], return_counts=True)

        .. versionadded:: 0.6.0
        """
        md = np.array(self.diploid_metadata, copy=False)
        deme_sizes = np.unique(md["deme"], return_counts=True)
        if as_dict is False:
            return deme_sizes
        return {i: j for i, j in zip(deme_sizes[0], deme_sizes[1])}

    def dump_tables_to_tskit(
        self, parameters: Optional[Dict] = None, *, destructive=False
    ):
        """
        Dump the population's TableCollection into
        an tskit TreeSequence

        :param parameters: The simulation parameters for the provenance table.
        :type parameters: None or dict

        :param destructive: If `True`, delete data held by the current instance.
        :type destructive: bool

        :rtype: tskit.TreeSequence

        .. warning::

            If `destructive` is `True`, further opertations on the
            current instance should be considered undefined behavior
            that may lead to a crash.

        .. versionchanged:: 0.8.2

            Added `parameters`.
            Generate provenance information for return value.
            The provenance information is validated using
            :func:`tskit.validate_provenance`, which may
            raise an exception.

        .. versionchanged:: 0.10.0

            Use tskit metadata schema.
            Mutation time is now stored in the tskit.MutationTable column.
            Origin time of mutations is part of the metadata.

        .. versionchanged:: 0.14.0

            Added `destructive` option.

        """
        return fwdpy11.tskit_tools._dump_tables_to_tskit._dump_tables_to_tskit(
            self, parameters, destructive=destructive
        )

    def dump_to_file(self, filename: str):
        """
        Write a population to a file in binary format.
        """
        self._dump_to_file(filename)

    def pickle_to_file(self, filename: IO):
        """
        Pickle the population to an open file.

        This function may be preferred over
        the direct pickling method because it uses less
        memory.  It is, however, slower.

        To read the population back in, you must call
        :func:`fwdpy11.DiploidPopulation.load_from_pickle_file`.

        :param f: A handle to an open file

        .. versionadded:: 0.3.0

        """
        self._pickle_to_file(filename)

    def sample_timepoints(
        self, include_alive=True
    ) -> Iterator[Tuple[int, np.ndarray, np.recarray]]:
        """
        Return an iterator over all sample time points.
        The iterator yields time, nodes, and metadata.

        :param include_alive: If True, include currently-alive individuals.

        .. versionadded :: 0.5.1

        .. versionchanged:: 0.5.3

           Monkey-patched into pybind11 class
        """
        amd = np.array(self.ancient_sample_metadata, copy=False)
        amd.flags.writeable = False
        times, utimes = self._get_times()
        for uti in utimes:
            mdidx = np.where(times == uti)[0]
            sample_nodes = amd["nodes"][mdidx].flatten()
            sample_nodes.flags.writeable = False
            mdslice = amd[mdidx][:]
            mdslice.flags.writeable = False
            yield uti, sample_nodes, mdslice

        if include_alive is True:
            md = np.array(self.diploid_metadata, copy=False)
            nodes = md["nodes"].flatten()
            md.flags.writeable = False
            nodes.flags.writeable = False
            yield self.generation, nodes, md



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

"""
This is basically a test of bad models.
"""

import unittest
import warnings

import numpy as np

import fwdpy11


def setup_and_run_model(pop, ddemog, simlen, recorder=None, seed=654321):
    pdict = {
        "nregions": [],
        "sregions": [],
        "recregions": [],
        "rates": (
            0,
            0,
            0,
        ),
        "gvalue": fwdpy11.Multiplicative(2.0),
        "demography": ddemog,
        "simlen": simlen,
    }
    params = fwdpy11.ModelParams(**pdict)
    rng = fwdpy11.GSLrng(seed)
    fwdpy11.evolvets(rng, pop, params, 100, recorder)


class TestNoDemographicEvents(unittest.TestCase):
    """
    For issue #594
    """

    @classmethod
    def setUpClass(self):
        self.demog = fwdpy11.DiscreteDemography()

    def test_no_demographic_events(self):
        _ = fwdpy11.DemographyDebugger([100], self.demog)


class TestEmptyEventLists(unittest.TestCase):
    """
    For issue #594
    """

    @classmethod
    def setUpClass(self):
        self.demog = fwdpy11.DiscreteDemography(
            mass_migrations=[],
            set_growth_rates=[],
            set_deme_sizes=[],
            set_selfing_rates=[],
            migmatrix=None,
            set_migration_rates=[],
        )

    def test_empty_event_lists(self):
        _ = fwdpy11.DemographyDebugger([100], self.demog)


class TestSingleDemeWithVariousMigrationMatrices(unittest.TestCase):
    """
    For issue #594
    """

    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1.0)

    def test_one_by_one_migration_matrix(self):
        """
        Will warn about the 1-by-1 matrix
        """
        demog = fwdpy11.DiscreteDemography(migmatrix=np.array([[1.0]]))
        with warnings.catch_warnings(record=True) as w:
            _ = fwdpy11.DemographyDebugger([100], demog)

    def test_two_by_two_migration_matrix(self):
        demog = fwdpy11.DiscreteDemography(migmatrix=np.array([[1.0, 0.0], [0.0, 1.0]]))
        with self.assertRaises(ValueError):
            _ = fwdpy11.DemographyDebugger([100], demog)


class TestBadMassMigrations(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1.0)

    def test_move_too_many(self):
        m = [
            fwdpy11.move_individuals(0, 0, 1, 0.5),
            fwdpy11.move_individuals(1, 0, 1, 1),
            fwdpy11.move_individuals(2, 0, 1, 1),
        ]
        d = fwdpy11.DiscreteDemography(mass_migrations=m)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)
        with self.assertRaises(fwdpy11.DemographyError):
            setup_and_run_model(self.pop, d, 5)

    def test_copy_from_empty_deme(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1), fwdpy11.copy_individuals(1, 0, 1, 1)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)
        with self.assertRaises(fwdpy11.DemographyError):
            setup_and_run_model(self.pop, d, 5)

    def test_failure_to_update_migration_matrix(self):
        M = np.zeros(4).reshape(2, 2)
        M[0, 0] = 1.0
        mm = [fwdpy11.move_individuals(when=10, source=0, destination=1, fraction=0.25)]
        d = fwdpy11.DiscreteDemography(migmatrix=M, mass_migrations=mm)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)


class TestDetectingExtinctions(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1.0)

    def test_no_valid_parents_V1(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_deme_sizes=s)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)

    def test_no_valid_parents_V2(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(1, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_deme_sizes=s)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)

    def test_no_valid_parents_V3(self):
        m = [fwdpy11.copy_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 0), fwdpy11.SetDemeSize(11, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_deme_sizes=s)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)

    def test_no_valid_parents_with_migration(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5] * 4).reshape(2, 2)
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_deme_sizes=s, migmatrix=M)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)

    def test_no_valid_parents_with_migration_V2(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5] * 4).reshape(2, 2)
        cM = [
            fwdpy11.SetMigrationRates(
                when=0, deme=None, migrates=np.array([0, 1, 0, 1]).reshape(2, 2)
            )
        ]
        d = fwdpy11.DiscreteDemography(
            mass_migrations=m, set_deme_sizes=s, set_migration_rates=cM, migmatrix=M
        )
        try:
            fwdpy11.DemographyDebugger(self.pop, d)
        except Exception:
            self.fail("unexpected exception")
        setup_and_run_model(self.pop, d, 5)

    def test_no_valid_parents_with_migration_V2_migration_rescue(self):
        """
        This is a convoluted model:
        0 moves 100% to 1. Thus, there are no parents anymore for 0.
        But, this is okay because the new deme 1 will provide parents
        back into 0, thus "un-extincting" 0.  Then, in the next
        generation, each deme gets 50/50 of its ancestry from each
        deme.
        """
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5] * 4).reshape(2, 2)
        cM = [
            fwdpy11.SetMigrationRates(0, None, np.array([0, 1, 0, 1]).reshape(2, 2)),
            fwdpy11.SetMigrationRates(1, None, np.array([0.5] * 4).reshape(2, 2)),
        ]
        d = fwdpy11.DiscreteDemography(
            mass_migrations=m, set_deme_sizes=s, set_migration_rates=cM, migmatrix=M
        )
        try:
            fwdpy11.DemographyDebugger(self.pop, d)
        except Exception:
            self.fail("unexpected exception")
        setup_and_run_model(self.pop, d, 5)

    def test_no_valid_parents_with_migration_V3(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5] * 4).reshape(2, 2)
        d = fwdpy11.DiscreteDemography(
            mass_migrations=m,
            set_deme_sizes=s,
            migmatrix=fwdpy11.MigrationMatrix(M, True),
        )
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)
        with self.assertRaises(fwdpy11.DemographyError):
            setup_and_run_model(self.pop, d, 5)


class TestBadModels(unittest.TestCase):
    def test_github_issue_544(self):
        pop = fwdpy11.DiploidPopulation(100, 1)
        mass_migrations = [
            fwdpy11.move_individuals(when=100, source=0, destination=1, fraction=0.5)
        ]
        m = 1e-3
        migmatrix = np.array([1.0 - m, m, m, 1.0 - m]).reshape(2, 2)
        bad_dmodel = fwdpy11.DiscreteDemography(
            mass_migrations=mass_migrations, migmatrix=migmatrix
        )
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger([100], bad_dmodel)

        pdict = {
            "nregions": [],
            "sregions": [],
            "recregions": [],
            "gvalue": fwdpy11.Multiplicative(1.0),
            "rates": [0, 0, 0],
            "demography": bad_dmodel,
            "simlen": 150,
        }

        rng = fwdpy11.GSLrng(1010)
        mp = fwdpy11.ModelParams(**pdict)
        with self.assertRaises(fwdpy11.DemographyError):
            fwdpy11.evolvets(rng, pop, mp, 100)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2019 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import unittest

import numpy as np

import fwdpy11


def setup_pdict(demog, simlen):
    nregions = []
    sregions = []
    recregions = []

    pdict = {
        "nregions": nregions,
        "sregions": sregions,
        "recregions": recregions,
        "rates": (0, 0, None),
        "gvalue": fwdpy11.Multiplicative(2.0),
        "demography": demog,
        "simlen": simlen,
        "prune_selected": True,
    }
    return pdict


def setup_pop_rng(N=100, seed=42):
    pop = fwdpy11.DiploidPopulation(N, 1.0)
    rng = fwdpy11.GSLrng(seed)
    return pop, rng


class TestTwoDemeIMModel(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.Nanc = 100
        self.N0, self.N1 = 1.1, 3.0
        from fwdpy11.demographic_models.IM import two_deme_IM

        self.model = two_deme_IM(
            self.Nanc, 0.1, 0.7, (self.N0, self.N1), (1e-2, 0.25), burnin=1.0
        )
        self.pop, self.rng = setup_pop_rng(self.Nanc)

    def test_complete_sim(self):
        pdict = setup_pdict(
            self.model.model,
            self.model.metadata.split_time + self.model.metadata.gens_post_split,
        )
        params = fwdpy11.ModelParams(**pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 10)
        self.assertEqual(self.pop.generation, params.simlen)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], np.rint(self.N0 * self.Nanc).astype(int))
        self.assertEqual(deme_sizes[1], np.rint(self.N1 * self.Nanc).astype(int))

    def test_evolve_in_two_steps(self):
        pdict = setup_pdict(self.model.model, self.model.metadata.split_time)
        params = fwdpy11.ModelParams(**pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 10)
        self.assertEqual(self.pop.generation, self.model.metadata.split_time)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], int((1.0 - 0.7) * self.Nanc))

        pdict["simlen"] = self.model.metadata.gens_post_split
        params = fwdpy11.ModelParams(**pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 10)
        self.assertEqual(
            self.pop.generation,
            self.model.metadata.split_time + self.model.metadata.gens_post_split,
        )
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], np.rint(self.N0 * self.Nanc).astype(int))
        self.assertEqual(deme_sizes[1], np.rint(self.N1 * self.Nanc).astype(int))

    def test_evolve_in_two_steps_restart_with_two_demes(self):
        deltat = 2
        pdict = setup_pdict(self.model.model, self.model.metadata.split_time + deltat)
        params = fwdpy11.ModelParams(**pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 10)
        self.assertEqual(self.pop.generation, self.model.metadata.split_time + deltat)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertTrue(0 in deme_sizes)
        self.assertTrue(1 in deme_sizes)

        pdict["simlen"] = self.model.metadata.gens_post_split - deltat
        params = fwdpy11.ModelParams(**pdict)
        fwdpy11.evolvets(
            self.rng, self.pop, params, 10, check_demographic_event_timings=False
        )
        self.assertEqual(
            self.pop.generation,
            self.model.metadata.split_time + self.model.metadata.gens_post_split,
        )
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], np.rint(self.N0 * self.Nanc).astype(int))
        self.assertEqual(deme_sizes[1], np.rint(self.N1 * self.Nanc).astype(int))


class TestTwoDemeIMModelVariousInitMethods(unittest.TestCase):
    def test_init_migrates_tuple(self):
        from fwdpy11.demographic_models.IM import two_deme_IM

        model = two_deme_IM(1000, 0.1, 0.7, (1.1, 2.7), (1e-2, 0.25), burnin=1.0)
        self.assertEqual(model, model)

    def test_init_migrates_list(self):
        from fwdpy11.demographic_models.IM import two_deme_IM

        model = two_deme_IM(1000, 0.1, 0.7, (1.1, 2.7), [1e-2, 0.25], burnin=1.0)
        self.assertEqual(model, model)

    def test_init_migrates_numpy(self):
        from fwdpy11.demographic_models.IM import two_deme_IM

        model = two_deme_IM(
            1000, 0.1, 0.7, (1.1, 2.7), np.array([1e-2, 0.25]), burnin=1.0
        )
        self.assertEqual(model, model)


@unittest.skipIf(
    "RUN_EXPENSIVE_TESTS" not in os.environ or os.environ["RUN_EXPENSIVE_TESTS"] != "1",
    "Expensive test.",
)
class TestTennessenModel(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from fwdpy11.demographic_models.human import tennessen

        self.demog = tennessen(0)
        self.simlen = self.demog.metadata["simlen"]
        self.Nref = self.demog.metadata["Nref"]
        self.pop = fwdpy11.DiploidPopulation(self.Nref, 1.0)
        self.pdict = setup_pdict(self.demog, self.simlen)
        self.params = fwdpy11.ModelParams(**self.pdict)
        self.rng = fwdpy11.GSLrng(915153)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)

    def test_generation(self):
        self.assertEqual(self.pop.generation, self.simlen)

    def test_final_deme_sizes(self):
        md = np.array(self.pop.diploid_metadata, copy=False)
        deme_sizes = np.unique(md["deme"], return_counts=True)
        expected_deme_sizes = [420000, 512000]
        for i, j in zip(deme_sizes[1], expected_deme_sizes):
            self.assertEqual(i, j)


@unittest.skipIf(
    "RUN_EXPENSIVE_TESTS" not in os.environ or os.environ["RUN_EXPENSIVE_TESTS"] != "1",
    "Expensive test.",
)
class TestTennessenModelV1(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from fwdpy11.demographic_models.human import tennessen

        self.demog = tennessen(0, fwdpy11.demographic_models.human.TennessenModel.V1)
        self.simlen = self.demog.metadata["simlen"]
        self.Nref = self.demog.metadata["Nref"]
        self.pop = fwdpy11.DiploidPopulation(self.Nref, 1.0)
        self.pdict = setup_pdict(self.demog, self.simlen)
        self.params = fwdpy11.ModelParams(**self.pdict)
        self.rng = fwdpy11.GSLrng(915153)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)

    def test_generation(self):
        self.assertEqual(self.pop.generation, self.simlen)

    def test_final_deme_sizes(self):
        md = np.array(self.pop.diploid_metadata, copy=False)
        deme_sizes = np.unique(md["deme"], return_counts=True)
        expected_deme_sizes = [423125, 501425]
        for i, j in zip(deme_sizes[1], expected_deme_sizes):
            self.assertEqual(i, j)


@unittest.skipIf(
    "RUN_EXPENSIVE_TESTS" not in os.environ or os.environ["RUN_EXPENSIVE_TESTS"] != "1",
    "Expensive test.",
)
class TestJouganousThreeDemeModel(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from fwdpy11.demographic_models.human import jouganous_three_deme

        self.demog = jouganous_three_deme(0)
        self.simlen = self.demog.metadata["simlen"]
        self.Nref = self.demog.metadata["Nref"]
        self.pop = fwdpy11.DiploidPopulation(self.Nref, 1.0)
        self.pdict = setup_pdict(self.demog, self.simlen)
        self.params = fwdpy11.ModelParams(**self.pdict)
        self.rng = fwdpy11.GSLrng(915153)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)

    def test_generation(self):
        self.assertEqual(self.pop.generation, self.simlen)

    def test_final_deme_sizes(self):
        md = np.array(self.pop.diploid_metadata, copy=False)
        deme_sizes = np.unique(md["deme"], return_counts=True)
        expected_deme_sizes = [23721, 39611, 83681]
        for i, j in zip(deme_sizes[1], expected_deme_sizes):
            self.assertEqual(i, j)


if __name__ == "__main__":
    unittest.main()



#
# Copyright(C) 2017 Kevin Thornton < krthornt @uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software : you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.If not, see < http: //www.gnu.org/licenses/>.
#

from typing import Callable, Optional

import fwdpy11

from ._fwdpy11 import GSLrng, SampleRecorder
from ._types import DiploidPopulation, ModelParams


def _validate_event_timings(demography: fwdpy11.DiscreteDemography, generation: int):
    too_early = []
    for i in demography._timed_events():
        if i is not None:
            for j in i:
                if j.when < generation:
                    too_early.append(j)
    if len(too_early) > 0:
        import warnings

        msg = "The following demographic events are registered to occur "
        msg += f"before the current generation, which is {generation}: {too_early}"
        warnings.warn(msg)


def evolvets(
    rng: GSLrng,
    pop: DiploidPopulation,
    params: ModelParams,
    simplification_interval: int,
    recorder: Optional[Callable[[DiploidPopulation, SampleRecorder], None]] = None,
    *,
    post_simplification_recorder: Optional[Callable[[DiploidPopulation], None]] = None,
    suppress_table_indexing: bool = False,
    record_gvalue_matrix: bool = False,
    stopping_criterion: Optional[Callable[[DiploidPopulation, bool], bool]] = None,
    track_mutation_counts: bool = False,
    remove_extinct_variants: bool = True,
    preserve_first_generation: bool = False,
    check_demographic_event_timings: bool = True,
):
    """
    Evolve a population with tree sequence recording

    :param rng: random number generator
    :type rng: :class:`fwdpy11.GSLrng`
    :param pop: A population
    :type pop: :class:`fwdpy11.DiploidPopulation`
    :param params: simulation parameters
    :type params: :class:`fwdpy11.ModelParams`
    :param simplification_interval: Number of generations between simplifications.
    :type simplification_interval: int
    :param recorder: (None) A temporal sampler/data recorder.
    :type recorder: typing.Callable
    :param post_simplification_recorder: (None) A temporal sampler
    :type post_simplification_recorder: typing.Callable
    :param suppress_table_indexing: (False) Prevents edge table indexing until
                                    end of simulation
    :type suppress_table_indexing: bool
    :param record_gvalue_matrix: (False) Whether to record genetic values into
                                 :attr:`fwdpy11.PopulationBase.genetic_values`.
    :type record_gvalue_matrix: bool
    :param preserve_first_generation: (False) Whether to record generation 0 as
                                      ancient samples. Must be `True` for
                                      tree sequence "recapitation". See
                                      :ref:`recapitation`.
    :type preserve_first_generation: bool
    :param check_demographic_event_timings: (True) If ``True``, then issue
                                            warnings if demographic events
                                            will occur prior to the current
                                            generation of the population.
    :type check_demographic_event_timings: bool

    The recording of genetic values into :attr:`fwdpy11.PopulationBase.genetic_values`
    is suppressed by default.  First, it is redundant with
    :attr:`fwdpy11.DiploidMetadata.g` for the common case of mutational effects on a
    single trait. Second, we save some memory by not tracking these matrices.
    However, it is useful to track these data for some
    cases when simulating multivariate mutational effects (pleiotropy).

    .. note::
        If recorder is None,
        then :class:`fwdpy11.NoAncientSamples` will be used.

        If post_simplification_recorder is None, then
        :class:`fwdpy11.RecordNothing` will be used.

    .. versionchanged:: 0.5.2

        Added post_simplification_recorder.

    .. versionchanged:: 0.7.1

        Added preserve_first_generation.

    .. versionchanged:: 0.8.0

        Update to refactored ModelParams.
        Added ``check_demographic_event_timings``.

    """
    if recorder is None:
        from ._fwdpy11 import NoAncientSamples

        recorder = NoAncientSamples()

    if post_simplification_recorder is None:
        from ._fwdpy11 import RecordNothing

        post_simplification_recorder = RecordNothing()
        reset_treeseqs_after_simplify = False
    else:
        reset_treeseqs_after_simplify = True

    if stopping_criterion is None:
        from ._fwdpy11 import _no_stopping

        stopping_criterion = _no_stopping

    from ._fwdpy11 import (MutationRegions, dispatch_create_GeneticMap,
                           evolve_with_tree_sequences)

    try:
        # DemographicModelDetails ?
        demographic_model = params.demography.model
    except AttributeError:
        # No, assume it is a DiscreteDemography
        demographic_model = params.demography
    except Exception as e:
        raise e

    if check_demographic_event_timings:
        _validate_event_timings(demographic_model, pop.generation)

    pneutral = 0.0
    if params.rates.neutral_mutation_rate + params.rates.selected_mutation_rate > 0.0:
        pneutral = params.rates.neutral_mutation_rate / (
            params.rates.neutral_mutation_rate + params.rates.selected_mutation_rate
        )
    mm = MutationRegions.create(pneutral, params.nregions, params.sregions)
    rm = dispatch_create_GeneticMap(params.rates.recombination_rate, params.recregions)

    from ._fwdpy11 import SampleRecorder

    sr = SampleRecorder()
    from ._fwdpy11 import _dgvalue_pointer_vector

    gvpointers = _dgvalue_pointer_vector(params.gvalue)
    evolve_with_tree_sequences(
        rng,
        pop,
        sr,
        simplification_interval,
        demographic_model,
        params.simlen,
        params.rates.neutral_mutation_rate,
        params.rates.selected_mutation_rate,
        mm,
        rm,
        gvpointers,
        recorder,
        stopping_criterion,
        params.prune_selected is False,
        suppress_table_indexing,
        record_gvalue_matrix,
        track_mutation_counts,
        remove_extinct_variants,
        reset_treeseqs_after_simplify,
        preserve_first_generation,
        post_simplification_recorder,
    )



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import pickle
import unittest

import numpy as np

import fwdpy11


class TestMoveOrCopyIndividuals(unittest.TestCase):
    """
    Tests functions for object construction and
    object pickling
    """

    def test_move(self):
        m = fwdpy11.move_individuals(0, 2, 1, 0.1)
        self.assertEqual(m.when, 0)
        self.assertEqual(m.source, 2)
        self.assertEqual(m.destination, 1)
        self.assertEqual(m.fraction, 0.1)
        self.assertEqual(m.move_individuals, True)
        self.assertEqual(m.resets_growth_rate, True)

        mp = pickle.dumps(m, -1)
        ump = pickle.loads(mp)
        self.assertEqual(ump.when, 0)
        self.assertEqual(ump.source, 2)
        self.assertEqual(ump.destination, 1)
        self.assertEqual(ump.fraction, 0.1)
        self.assertEqual(ump.move_individuals, True)
        self.assertEqual(ump.resets_growth_rate, True)
        self.assertTrue(m == ump)

    def test_copy(self):
        m = fwdpy11.copy_individuals(0, 2, 1, 0.1)
        self.assertEqual(m.when, 0)
        self.assertEqual(m.source, 2)
        self.assertEqual(m.destination, 1)
        self.assertEqual(m.fraction, 0.1)
        self.assertEqual(m.move_individuals, False)

        mp = pickle.dumps(m, -1)
        ump = pickle.loads(mp)
        self.assertEqual(ump.when, 0)
        self.assertEqual(ump.source, 2)
        self.assertEqual(ump.destination, 1)
        self.assertEqual(ump.fraction, 0.1)
        self.assertEqual(ump.move_individuals, False)
        self.assertTrue(m == ump)


class TestSetDemeSize(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.ddsc = fwdpy11.SetDemeSize(0, 1, 3)

    def test_property_access(self):
        self.assertEqual(self.ddsc.when, 0)
        self.assertEqual(self.ddsc.deme, 1)
        self.assertEqual(self.ddsc.new_size, 3)
        self.assertEqual(self.ddsc.resets_growth_rate, True)

    def test_pickle(self):
        p = pickle.dumps(self.ddsc, -1)
        up = pickle.loads(p)
        self.assertEqual(self.ddsc.when, up.when)
        self.assertEqual(self.ddsc.deme, up.deme)
        self.assertEqual(self.ddsc.new_size, up.new_size)

    def test_invalid_deme(self):
        with self.assertRaises(ValueError):
            self.dpsc = fwdpy11.SetDemeSize(0, -1, 3)


class TestSetExponentialGrowth(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.seg = fwdpy11.SetExponentialGrowth(0, 1, 0.1)

    def test_property_access(self):
        self.assertEqual(self.seg.when, 0)
        self.assertEqual(self.seg.deme, 1)
        self.assertEqual(self.seg.G, 0.1)

    def test_pickle(self):
        p = pickle.dumps(self.seg, -1)
        up = pickle.loads(p)
        self.assertEqual(self.seg.when, up.when)
        self.assertEqual(self.seg.deme, up.deme)
        self.assertEqual(self.seg.G, up.G)

    def test_invalid_deme(self):
        with self.assertRaises(ValueError):
            self.seg = fwdpy11.SetExponentialGrowth(0, -1, 0.1)

    def test_invalid_growth_rate(self):
        with self.assertRaises(ValueError):
            self.seg = fwdpy11.SetExponentialGrowth(0, 1, np.nan)


class TestMigrationMatrix(unittest.TestCase):
    def testSimpleInit(self):
        fwdpy11.MigrationMatrix(np.identity(2))

    def testShape(self):
        m = fwdpy11.MigrationMatrix(np.identity(2))
        self.assertEqual(m.shape, (2, 2))

    def testPickle(self):
        m = fwdpy11.MigrationMatrix(np.identity(2))
        p = pickle.dumps(m, -1)
        up = pickle.loads(p)
        self.assertEqual(m, up)
        self.assertEqual(up.shape, (2, 2))
        self.assertTrue(np.array_equal(up.M, np.identity(2)))

    @unittest.skip("obsolete, and a bad idea anyways")
    def testSettingRates(self):
        m = fwdpy11.MigrationMatrix(np.identity(2))
        m._set_migration_rates(0, [0.5, 0.5])
        m._set_migration_rates(1, np.array([0.01, 0.99]))
        self.assertTrue(
            np.array_equal(m.M, np.array([0.5, 0.5, 0.01, 0.99]).reshape(2, 2))
        )

        m._set_migration_rates(np.array([0, 1, 2, 3]).reshape(2, 2))

        try:
            m._set_migration_rates(0, [2.5, 0.5])
        except:  # NOQA
            self.fail("unexpected exception")

        with self.assertRaises(ValueError):
            # Fails because deme index out of range
            m._set_migration_rates(2, [0.5, 0.5])

    def test_non_square_input(self):
        with self.assertRaises(ValueError):
            fwdpy11.MigrationMatrix(np.array([2.0] * 2))

        with self.assertRaises(ValueError):
            """This matrix is also invalid b/c
            rows don't sum to 1, but we expect
            a failure before that check"""
            fwdpy11.MigrationMatrix(np.array([i for i in range(6)]).reshape(2, 3))

    def test_weights_greater_than_one(self):
        with self.assertRaises(ValueError):
            fwdpy11.MigrationMatrix(np.array([2.0] * 4).reshape(2, 2), False)


class TestSetMigrationRates(unittest.TestCase):
    def test_init_from_list(self):
        m = fwdpy11.SetMigrationRates(0, 0, [0, 1, 0])
        self.assertEqual(m.when, 0)
        self.assertEqual(m.deme, 0)
        self.assertTrue(np.array_equal(m.migrates, np.array([0, 1, 0])))

    def test_init_from_numpy(self):
        m = fwdpy11.SetMigrationRates(0, 0, np.array([0, 1, 0]))
        self.assertEqual(m.when, 0)
        self.assertEqual(m.deme, 0)
        self.assertTrue(np.array_equal(m.migrates, np.array([0, 1, 0])))

    def test_reset_entire_matrix(self):
        m = fwdpy11.SetMigrationRates(3, None, np.identity(3))
        self.assertTrue(np.array_equal(m.migrates, np.identity(3)))

    def test_reset_entire_matrix_bad_inputs(self):
        with self.assertRaises(ValueError):
            fwdpy11.SetMigrationRates(3, None, np.arange(4).reshape(2, 2))

    def test_pickle(self):
        m = fwdpy11.SetMigrationRates(0, 0, np.array([0, 1, 0]))
        p = pickle.dumps(m, -1)
        up = pickle.loads(p)
        self.assertEqual(m.when, up.when)
        self.assertEqual(m.deme, up.deme)
        self.assertEqual(m.migrates.tolist(), up.migrates.tolist())


class TestDiscreteDemographyInitialization(unittest.TestCase):
    @unittest.skip("obsolete: UI may need rethinking")
    def test_init_migmatrix_with_tuple(self):
        mm = np.array([0.3, 0.7, 0.7, 0.3]).reshape(2, 2)
        try:
            d = fwdpy11.DiscreteDemography(migmatrix=mm)
            self.assertEqual(d.migmatrix.scaled, False)
        except:  # NOQA
            self.fail("unexpected exception")
        try:
            d = fwdpy11.DiscreteDemography(migmatrix=((mm, True)))
            self.assertEqual(d.migmatrix.scaled, True)
        except:  # NOQA
            self.fail("unexpected exception")
        try:
            d = fwdpy11.DiscreteDemography(migmatrix=((mm, False)))
            self.assertEqual(d.migmatrix.scaled, False)
        except:  # NOQA
            self.fail("unexpected exception")

    def test_move_or_copy(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        dd = fwdpy11.DiscreteDemography(m)
        self.assertTrue(dd.migmatrix is None)
        with self.assertRaises(AttributeError):
            del self.migmatrix

    def test_setting_growth(self):
        g = [fwdpy11.SetExponentialGrowth(0, 1, 0.3)]
        fwdpy11.DiscreteDemography(set_growth_rates=g)

    def test_setting_size_changes(self):
        c = [fwdpy11.SetDemeSize(0, 1, 15151)]
        fwdpy11.DiscreteDemography(set_deme_sizes=c)

    def test_init_from_numpy(self):
        N = 1000
        popsizes = np.array(
            [N] * 10 * N + [int(0.1 * N)] * int(0.1 * N) + [2 * N] * 100,
            dtype=np.uint32,
        )
        demography = fwdpy11.DiscreteDemography(set_deme_sizes=popsizes)
        self.assertEqual(len(np.unique(popsizes)), len(demography.set_deme_sizes))

    def test_init_with_two_growth_changes_same_generation(self):
        g = [fwdpy11.SetExponentialGrowth(0, 1, 0.3)] * 2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(set_growth_rates=g)

    def test_init_with_two_selfing_changes_same_generation(self):
        g = [fwdpy11.SetSelfingRate(0, 1, 0.3)] * 2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(set_selfing_rates=g)

    def test_init_with_two_deme_size_changes_same_generation(self):
        c = [fwdpy11.SetDemeSize(0, 1, 15151)] * 2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(set_deme_sizes=c)

    def test_init_with_two_mass_moves_same_generation(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5)] * 2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(m)

    def test_init_with_two_mass_copies_same_generation(self):
        m = [fwdpy11.copy_individuals(0, 0, 1, 0.5)] * 2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(m)

    def test_init_with_one_mass_move_and_one_mass_copy_same_generation(self):
        # Internally, the input data will get sorted
        # so that copies happen before moves.
        m = [
            fwdpy11.move_individuals(0, 0, 1, 0.5),
            fwdpy11.copy_individuals(0, 0, 1, 0.5),
        ]
        try:
            d = fwdpy11.DiscreteDemography(m)
            self.assertEqual(d.mass_migrations[0].move_individuals, False)
            self.assertEqual(d.mass_migrations[1].move_individuals, True)
        except:  # NOQA
            self.fail("unexpected exception")

    def test_move_too_many_individuals_from_same_deme(self):
        # Attempt to move 125% of deme 0
        m = [
            fwdpy11.move_individuals(0, 0, 1, 0.5),
            fwdpy11.move_individuals(0, 0, 2, 0.75),
        ]
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(m)

    def test_move_too_many_individuals_from_same_deme_with_copy_in_middle(self):
        # Attempt to move 125% of deme 0
        # This is a test that input data get sorted so
        # that copies are before moves, which will allow
        # detecting the invalid cumulative move attempt
        m = [
            fwdpy11.move_individuals(0, 0, 1, 0.5),
            fwdpy11.copy_individuals(0, 0, 1, 0.5),
            fwdpy11.move_individuals(0, 0, 2, 0.75),
        ]
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(m)

    def test_with_migration_matrix(self):
        mm = np.array([0.72, 0.28, 0.3, 0.7]).reshape(2, 2)
        d = fwdpy11.DiscreteDemography(migmatrix=mm)
        p = pickle.dumps(d, -1)
        up = pickle.loads(p)
        self.assertTrue(np.array_equal(up.migmatrix.M, mm))
        with self.assertRaises(AttributeError):
            del d.migmatrix

    @unittest.skip("obsolete")
    def test_identity_matrix_conversion_to_None(self):
        mm = np.identity(5)
        d = fwdpy11.DiscreteDemography(migmatrix=mm)
        self.assertTrue(d.migmatrix is None)

    def test_identity_matrix_with_migrate_changes(self):
        mm = np.identity(5)
        cm = [fwdpy11.SetMigrationRates(0, 3, [0.2] * 5)]
        d = fwdpy11.DiscreteDemography(migmatrix=mm, set_migration_rates=cm)
        self.assertTrue(d.migmatrix is not None)


class TestEventSorting(unittest.TestCase):
    """
    Doesn't test all possibilities, esp. for MassMigration events
    """

    def test_events_are_sorted(self):
        d = fwdpy11.DiscreteDemography(
            mass_migrations=[
                fwdpy11.MassMigration(when=10, source=0, destination=1, fraction=0.25),
                fwdpy11.MassMigration(when=1, source=1, destination=0, fraction=0.25),
            ],
            set_deme_sizes=[
                fwdpy11.SetDemeSize(when=10, deme=0, new_size=11),
                fwdpy11.SetDemeSize(when=3, deme=1, new_size=100),
            ],
            set_growth_rates=[
                fwdpy11.SetExponentialGrowth(when=10, deme=0, G=0.25),
                fwdpy11.SetExponentialGrowth(when=1, deme=0, G=fwdpy11.NOGROWTH),
            ],
            set_selfing_rates=[
                fwdpy11.SetSelfingRate(when=303, deme=0, S=0.1),
                fwdpy11.SetSelfingRate(when=1, deme=0, S=1.0),
            ],
            set_migration_rates=[
                fwdpy11.SetMigrationRates(when=10, deme=0, migrates=[0.5, 0.5]),
                fwdpy11.SetMigrationRates(when=8, deme=0, migrates=[1, 0.0]),
            ],
            migmatrix=np.eye(2),
        )
        for i in [
            d.mass_migrations,
            d.set_deme_sizes,
            d.set_growth_rates,
            d.set_selfing_rates,
            d.set_migration_rates,
        ]:
            self.assertTrue(i == sorted(i, key=lambda x: x.when))


class TestPickling(unittest.TestCase):
    """
    Test pickling a couple of models that use all of the features.
    """

    def test_pickle_IM_with_selfing(self):
        import fwdpy11.demographic_models.IM as IM

        temp = IM.two_deme_IM(1000, 100, 0.25, [0.5, 3], [0.1, 0.25]).model.asdict()
        temp["set_selfing_rates"] = [fwdpy11.SetSelfingRate(0, 0, 0.1)]
        im = fwdpy11.DiscreteDemography(**temp)
        pi = pickle.dumps(im)
        up = pickle.loads(pi)
        self.assertEqual(im, up)

    def test_pickle_IM_with_whole_migmatrix_reset(self):
        import fwdpy11.demographic_models.IM as IM

        temp = IM.two_deme_IM(1000, 100, 0.25, [0.5, 3], [0.1, 0.25]).model.asdict()
        temp["set_migration_rates"].append(
            fwdpy11.SetMigrationRates(10000 + 10, None, np.eye(2))
        )
        im = fwdpy11.DiscreteDemography(**temp)
        pi = pickle.dumps(im)
        up = pickle.loads(pi)
        self.assertEqual(im, up)

    def test_pickle_tennessen(self):
        import fwdpy11.demographic_models.human as human

        tennessen = human.tennessen()
        pt = pickle.dumps(tennessen)
        up = pickle.loads(pt)
        self.assertEqual(tennessen, up)

    def test_pickle_model_with_scaled_migration_matrix(self):
        m = fwdpy11.DiscreteDemography(migmatrix=(np.eye(2), True))
        p = pickle.dumps(m)
        up = pickle.loads(p)
        self.assertEqual(m, up)


if __name__ == "__main__":
    unittest.main()



"""
This module provided pre-calculated demographic models
for human populations.
"""
import enum

import attr
import numpy as np

import fwdpy11

from .demographic_model_details import (DemographicModelCitation,
                                        DemographicModelDetails)


class TennessenModel(enum.Enum):
    """
    Enumeration type declaring alternative implementations
    of the Tennessen et al. (2012) model. See
    :func:`fwdpy11.demographic_models.human.tennessen` for
    details.
    """

    V0 = 0
    """
    Default version of the model
    """
    V1 = 1
    """
    Alternate version of the model
    """


@attr.s()
class _ModelParameterValidator(object):
    """ Make sure that we get an actual int """

    burnin = attr.ib(type=int, kw_only=True, validator=attr.validators.instance_of(int))

    @burnin.validator
    def validate_burnin(self, attribute, value):
        if value < 0:
            raise ValueError("burnin must be >= 0")


def tennessen(burnin: int = 20, model_version: TennessenModel = TennessenModel.V0):
    """
    Generate parameters for the demographic model described in:

    Tennessen, Jacob A., Abigail W. Bigham, Timothy D. O’Connor, Wenqing Fu,
    Eimear E. Kenny, Simon Gravel, Sean McGee, et al. 2012.
    “Evolution and Functional Impact of Rare Coding Variation
    from Deep Sequencing of Human Exomes.” Science 337 (6090): 64–69.

    :param burnin: Burn-in time, as a multiplier of the ancestral population size.
                   Defaults to 20.
    :type param: int
    :param model_version: Specify which "flavor" of the model to run.  See below.
    :type model_version: fwdpy11.demographic_models.human.TennessenModel

    :returns: The demographic model
    :rtype: fwdpy11.demographic_models.DemographicModelDetails

    The ``metadata`` field of the return value contains information about
    the simulation length, ancestral population size (``Nref``), etc.,
    and the mapping of integer values to deme names.

    When this model is run, deme ``0`` corresponds to ``African`` and deme
    ``1`` corresponds to Eurasian.

    Due to uncertainty in the literature over the details of this model,
    we provide two slightly different implementations.  The first
    implementation is the default, and is specified by
    :attr:`fwdpy11.demographic_models.human.TennessenModel.V0`.  For this
    model, the final deme sizes are those shown in Figure 3 of the
    Tennessen et al. (2012) paper.
    If :attr:`fwdpy11.demographic_models.human.TennessenModel.V1` is
    used instead, then the final sizes are obtained by calculating them
    forwards in time using a combination of the information from
    Tennessen et al. (2012) and Fu et al. 2013. However, due to having
    to convert from continuous to discrete time, etc., other definitions
    of this model are also possible.  Here, all deme sizes and times
    are obtained using :data:`numpy.rint`.


    .. note::

        This implementation is based on code provided by Aaron Ragsdale.

    .. versionadded:: 0.8.0
    """
    _ModelParameterValidator(burnin=burnin)
    Nref = 7310  # Ancestral population dize
    NAfr0 = 14474  # Initial size change
    NB = 1861  # Eurasian bottleneck size
    if model_version == TennessenModel.V0:
        NAfr = 420000  # Final African pop'n size
    elif model_version == TennessenModel.V1:
        NAfr = 423125  # Final African pop'n size
    else:
        raise ValueError("invalid model_version")

    NEur0 = 1032  # Eurasian second bottleneck size
    if model_version == TennessenModel.V0:
        NEur1 = 9237  # Eurasian size as rapid growth starts
    elif model_version == TennessenModel.V1:
        NEur1 = 9279  # Eurasian size as rapid growth starts
    else:
        raise ValueError("invalid model_version")

    if model_version == TennessenModel.V0:
        NEur = 512000  # Final Eurasian pop'n size
    elif model_version == TennessenModel.V1:
        NEur = 501425  # Final Eurasian pop'n size
    else:
        raise ValueError("invalid model_version")
    T_Af = 148000
    T_B = 51000
    T_Eu_As = 23000
    T_accel = 5115
    mB = 15e-5
    mF = 2.5e-5
    generation_time = 25

    # List of demographic events:
    # keep track of size change, copying, and migration rate change events in
    # separate lists
    size_change = []
    copy = []
    mig_rates = []
    growth_rates = []

    # number of generations in epochs
    T0 = np.rint((T_Af - T_B) / generation_time).astype(int)  # pre-split
    # split to bottleneck, no growth
    T1 = np.rint((T_B - T_Eu_As) / generation_time).astype(int)
    T2 = np.rint((T_Eu_As - T_accel) / generation_time).astype(
        int
    )  # Eu growth with r_Eu0
    # accelerated growth in Af and Eu
    T3 = np.rint(T_accel / generation_time).astype(int)

    M_init = np.zeros(4).reshape(2, 2)
    M_init[0, 0] = 1
    mm = fwdpy11.MigrationMatrix(M_init)

    # burn in for 20*Ne generations
    gens_burn_in = burnin * Nref
    total_sim_length = gens_burn_in + T0 + T1 + T2 + T3

    # init: size change of common ancestral population
    size_change.append(fwdpy11.SetDemeSize(when=gens_burn_in, deme=0, new_size=NAfr0))

    # T0: mass migration, copy from A to Eu bottleneck population
    copy.append(
        fwdpy11.copy_individuals(
            when=gens_burn_in + T0, source=0, destination=1, fraction=NB / NAfr0,
        )
    )
    size_change.append(fwdpy11.SetDemeSize(when=gens_burn_in + T0, deme=1, new_size=NB))
    # at the same time, set migration rate between deme 0 and 1 to m_A_B
    mig_rates.append(fwdpy11.SetMigrationRates(gens_burn_in + T0, 0, [1 - mB, mB]))
    mig_rates.append(fwdpy11.SetMigrationRates(gens_burn_in + T0, 1, [mB, 1 - mB]))

    # T1: adjust size of Eu to Eu0 and set growth rate
    size_change.append(
        fwdpy11.SetDemeSize(when=gens_burn_in + T0 + T1, deme=1, new_size=NEur0)
    )
    r_Eur0 = (NEur1 / NEur0) ** (1 / T2) - 1
    growth_rates.append(
        fwdpy11.SetExponentialGrowth(when=gens_burn_in + T0 + T1, deme=1, G=1 + r_Eur0)
    )
    # set migration rates to contemporary rates
    mig_rates.append(fwdpy11.SetMigrationRates(gens_burn_in + T0 + T1, 0, [1 - mF, mF]))
    mig_rates.append(fwdpy11.SetMigrationRates(gens_burn_in + T0 + T1, 1, [mF, 1 - mF]))

    # T2: set growth rates to accelerated rates in both populations
    r_AfrF = (NAfr / NAfr0) ** (1 / T3) - 1
    r_EurF = (NEur / NEur1) ** (1 / T3) - 1
    growth_rates.append(
        fwdpy11.SetExponentialGrowth(
            when=gens_burn_in + T0 + T1 + T2, deme=0, G=1 + r_AfrF
        )
    )
    growth_rates.append(
        fwdpy11.SetExponentialGrowth(
            when=gens_burn_in + T0 + T1 + T2, deme=1, G=1 + r_EurF
        )
    )

    ddemog = fwdpy11.DiscreteDemography(
        mass_migrations=copy,
        set_deme_sizes=size_change,
        migmatrix=mm,
        set_migration_rates=mig_rates,
        set_growth_rates=growth_rates,
    )
    full_citation = (
        f"Tennessen, Jacob A., Abigail W. Bigham, Timothy D. O’Connor, "
        f"Wenqing Fu, Eimear E. Kenny, Simon Gravel, Sean McGee, et al. 2012. "
        f"“Evolution and Functional Impact of Rare Coding Variation from Deep "
        f"Sequencing of Human Exomes.” Science 337 (6090): 64–69. ",
        f"and Figure S5 of: ",
        f"Fu, Wenqing, Timothy D. O’Connor, Goo Jun, Hyun Min Kang, Goncalo Abecasis, "
        f"Suzanne M. Leal, Stacey Gabriel, et al. 2013. “Analysis of 6,515 "
        f"Exomes Reveals the Recent Origin of Most Human Protein-Coding Variants.” "
        f"Nature 493 (7431): 216–20",
    )

    full_citation_str = str().join([str(i) for i in full_citation])

    return DemographicModelDetails(
        model=ddemog,
        name="Tennessen et al. model of African and European demography.",
        source={"function": "fwdpy11.demographic_models.human.tennessen"},
        parameters={"burnin": burnin},
        citation=DemographicModelCitation(
            DOI="10.1126/science.1219240 and 10.1038/nature11690",
            full_citation=full_citation_str,
            metadata=None,
        ),
        metadata={
            "deme_labels": {0: "African", 1: "Eurasian"},
            "simlen": total_sim_length,
            "Nref": Nref,
        },
    )


def jouganous_three_deme(burnin: int = 20):
    """
    Generate parameters for the demographic model described in Table 2 of:

    Jouganous, Julien, Will Long, Aaron P. Ragsdale, and Simon Gravel. 2017.
    “Inferring the Joint Demographic History of Multiple Populations:
    Beyond the Diffusion Approximation.” Genetics 206 (3): 1549–67.

    :param burnin: Burn-in time, as a multiplier of the ancestral population size.
                   Defaults to 20.
    :type param: int
    :returns: The demographic model
    :rtype: fwdpy11.demographic_models.DemographicModelDetails

    The ``metadata`` field of the return value contains information about
    the simulation length, ancestral population size (``Nref``), etc.,
    and the mapping of integer values to deme names.

    .. versionadded:: 0.8.1

    """
    _ModelParameterValidator(burnin=burnin)

    NA = 11273  # Ancestral population size

    gens_burn_in = burnin * NA

    # TAF years ago, change the ancestral
    # deme size to the African deme size.
    TAF = 312000
    NAF = 23721

    # TB years ago, the ancestral population
    # splits into the population ancestral
    # to modern Africa and population B,
    # which is the ancestor of modern
    # Eurasian populations.
    TB = 125000
    NB = 3104

    # TAS_EU years ago, pop B splits into
    # EU (modern European) and AS (modern
    # Asian).
    TAS_EU = 42300
    NAS0 = 923
    NEU0 = 2271

    generation_time = 29
    # These are the CONTINUOUS growth
    # rates inferred in Jouganous et al.
    rAS = 0.00309  # units of 2NA
    rEU = 0.00196
    tg = TAS_EU / generation_time  # Time AS and EU grow exponentially, in generations

    # Get the discrete-time growth rates...
    NASF = NAS0 * np.exp(rAS * tg)
    NEUF = NEU0 * np.exp(rEU * tg)
    NASFi = np.rint(NASF)
    NEUFi = np.rint(NEUF)

    # ... which are these:
    dGAS = np.exp((np.log(NASFi) - np.log(NAS0)) / np.rint(tg))
    dGEU = np.exp((np.log(NEUFi) - np.log(NEU0)) / np.rint(tg))

    # Now, convert all the times from years into discrete
    # generations.

    TAFg = 0

    TBg = np.rint((TAF - TB) / generation_time).astype(int)
    TAS_EUg = np.rint((TB - TAS_EU) / generation_time).astype(int)
    total_sim_length = gens_burn_in + TAFg + TBg + TAS_EUg + np.rint(tg).astype(int)

    # Build up the events list
    deme_labels = {"AFR": 0, "B_CEU": 1, "CHB": 2}

    migmatrix = np.zeros(9).reshape(3, 3)
    migmatrix[0, 0] = 1.0
    change_migrates = []

    T = TAFg

    # Change size of AFR
    set_deme_sizes = [
        fwdpy11.SetDemeSize(
            when=gens_burn_in + T, deme=deme_labels["AFR"], new_size=NAF
        )
    ]

    T += TBg

    # "Bud" off deme B
    mass_migrations = [
        fwdpy11.copy_individuals(
            when=gens_burn_in + T,
            source=deme_labels["AFR"],
            destination=deme_labels["B_CEU"],
            fraction=NB / NAF,
        )
    ]

    mAF_B = 15.8e-5
    change_migrates.append(
        fwdpy11.SetMigrationRates(
            when=gens_burn_in + T,
            deme=deme_labels["AFR"],
            migrates=[1.0 - mAF_B, mAF_B, 0.0],
        )
    )
    change_migrates.append(
        fwdpy11.SetMigrationRates(
            when=gens_burn_in + T,
            deme=deme_labels["B_CEU"],
            migrates=[mAF_B, 1.0 - mAF_B, 0.0],
        )
    )

    # Split B into CEU and CHB ("European" and "Asian")
    # CEU inherits the deme label of B.
    T += TAS_EUg
    mass_migrations.append(
        fwdpy11.copy_individuals(
            when=gens_burn_in + T,
            source=deme_labels["B_CEU"],
            destination=deme_labels["CHB"],
            fraction=NAS0 / NB,
        )
    )

    set_deme_sizes.append(
        fwdpy11.SetDemeSize(
            when=gens_burn_in + T, deme=deme_labels["B_CEU"], new_size=NEU0
        )
    )

    set_growth_rates = [
        fwdpy11.SetExponentialGrowth(
            when=gens_burn_in + T, deme=deme_labels["B_CEU"], G=dGEU
        )
    ]

    set_growth_rates.append(
        fwdpy11.SetExponentialGrowth(
            when=gens_burn_in + T, deme=deme_labels["CHB"], G=dGAS
        )
    )

    mAF_CEU = 1.10e-5
    mAF_CHB = 0.48e-5
    mCEU_CHB = 4.19e-5

    change_migrates.append(
        fwdpy11.SetMigrationRates(
            when=gens_burn_in + T,
            deme=deme_labels["AFR"],
            migrates=[1.0 - (mAF_CEU + mAF_CHB), mAF_CEU, mAF_CHB],
        )
    )

    change_migrates.append(
        fwdpy11.SetMigrationRates(
            when=gens_burn_in + T,
            deme=deme_labels["B_CEU"],
            migrates=[mAF_CEU, 1.0 - (mAF_CEU + mCEU_CHB), mCEU_CHB],
        )
    )

    change_migrates.append(
        fwdpy11.SetMigrationRates(
            when=gens_burn_in + T,
            deme=deme_labels["CHB"],
            migrates=[mAF_CHB, mCEU_CHB, 1.0 - (mAF_CHB + mCEU_CHB)],
        )
    )

    ddemog = fwdpy11.DiscreteDemography(
        mass_migrations=mass_migrations,
        set_deme_sizes=set_deme_sizes,
        migmatrix=migmatrix,
        set_migration_rates=change_migrates,
        set_growth_rates=set_growth_rates,
    )

    full_citation = (
        f"Jouganous, Julien, Will Long, Aaron P. Ragsdale, and Simon Gravel. 2017."
        f" “Inferring the Joint Demographic History "
        f"of Multiple Populations: Beyond the Diffusion Approximation.”"
        f" Genetics 206 (3): 1549–67."
    )
    full_citation = str().join([str(i) for i in full_citation])

    return DemographicModelDetails(
        model=ddemog,
        name="Jouganous et al. 3-deme model.",
        source={"function": "fwdpy11.demographic_models.human.jouganous_three_deme"},
        parameters={"burnin": burnin},
        citation=DemographicModelCitation(
            DOI="10.1534/genetics.117.200493",
            full_citation=full_citation,
            metadata="Parameters from Table 2.",
        ),
        metadata={
            "deme_labels": {value: key for key, value in deme_labels.items()},
            "simlen": total_sim_length,
            "Nref": NA,
        },
    )



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

import scipy.stats

import fwdpy11
import sregion_cdf


class TestSregionFromCDF(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # Test everything w.r.to a value of 0.1
        # from the underlying cdf
        self.P = 0.1

    def get_deviate(self, dist):
        return sregion_cdf.from_mvnorm(0.0, self.P, dist)

    def test_ExpS(self):
        mean = 0.2
        e = fwdpy11.ExpS(0, 1, 1, mean)
        d = self.get_deviate(e)
        exp = scipy.stats.expon(0, mean)
        self.assertAlmostEqual(self.P, exp.cdf(d))

    def test_ExpS_with_scaling(self):
        mean = 0.2
        scaling = 1e3
        e = fwdpy11.ExpS(0, 1, 1, mean * scaling, scaling=scaling)
        d = self.get_deviate(e)
        exp = scipy.stats.expon(0, mean)
        self.assertAlmostEqual(self.P, exp.cdf(d))

    def test_GaussianS(self):
        sd = 0.333
        g = fwdpy11.GaussianS(0, 1, 1, sd)
        d = self.get_deviate(g)
        norm = scipy.stats.norm(0, sd)
        self.assertAlmostEqual(self.P, norm.cdf(d))

    def test_GaussianS_with_scaling(self):
        sd = 0.333
        scaling = 1e3
        g = fwdpy11.GaussianS(0, 1, 1, sd * scaling, scaling=scaling)
        d = self.get_deviate(g)
        norm = scipy.stats.norm(0, sd)
        self.assertAlmostEqual(self.P, norm.cdf(d))

    def test_GammaS(self):
        # a = shape
        # b = mean/shape
        mean = 5
        shape = 2.6
        g = fwdpy11.GammaS(0, 1, 1, mean=mean, shape_parameter=shape)
        d = self.get_deviate(g)
        gamma = scipy.stats.gamma(a=shape, scale=mean / shape)
        self.assertAlmostEqual(self.P, gamma.cdf(d))

    def test_GammaS_with_scaling(self):
        # a = shape
        # b = mean/shape
        scaling = 1e3
        mean = 5 * scaling
        shape = 2.6
        g = fwdpy11.GammaS(0, 1, 1, mean=mean, shape_parameter=shape, scaling=scaling)
        d = self.get_deviate(g)
        gamma = scipy.stats.gamma(a=shape, scale=mean / (scaling * shape))
        self.assertAlmostEqual(self.P, gamma.cdf(d))

    def test_UniformS(self):
        lo = 0.1
        hi = 0.47
        u = fwdpy11.UniformS(0, 1, 1, lo, hi)
        d = self.get_deviate(u)
        uniform = scipy.stats.uniform(loc=lo, scale=hi - lo)
        self.assertAlmostEqual(self.P, uniform.cdf(d))

    def test_UniformS_with_scaling(self):
        lo = 100
        hi = 200
        scaling = 1e3
        u = fwdpy11.UniformS(0, 1, 1, lo, hi, scaling=scaling)
        d = self.get_deviate(u)
        uniform = scipy.stats.uniform(loc=lo, scale=hi - lo)
        self.assertAlmostEqual(self.P, uniform.cdf(d * scaling))


if __name__ == "__main__":
    unittest.main()



from typing import (
    Dict,
    List,
    Optional,
    Union,
)
import numpy as np
import attr
import math
import sys
import copy
import itertools

import demes

from ..discrete_demography import (
    MassMigration,
    MigrationMatrix,
    SetDemeSize,
    SetExponentialGrowth,
    SetMigrationRates,
    SetSelfingRate,
    copy_individuals,
    move_individuals,
    DiscreteDemography,
)

from .. import class_decorators

from ..demographic_models import DemographicModelDetails, DemographicModelCitation

from .._demography import exponential_growth_rate

# TODO: need type hints for dg
def demography_from_demes(
    dg: Union[str, demes.Graph], burnin: int
) -> DemographicModelDetails:
    """
    The deme graph, dg, can be either a string or a resolved deme-graph.

    :param dg: A demes graph, either the object itself or a string for the YAML
        specified demography.
    :param int burnin: The factor for the burnin time, so that burn in occurs for
        burnin * N generations.

    .. versionadded:: 0.14.0
    """
    if isinstance(dg, str):
        g = demes.load(dg)
        source = {"demes_yaml_file": dg}
    else:
        g = dg
        source = None

    if not isinstance(burnin, int):
        raise ValueError("The burn in factor must be an integer")
    if burnin < 0:
        raise ValueError("Burn in factor must be non-negative")

    demography = _build_from_deme_graph(g, burnin, source)
    return demography


def _build_from_deme_graph(
    dg: demes.Graph, burnin: int, source: Optional[Dict] = None
) -> DemographicModelDetails:
    """
    The workhorse.
    """
    # dg must be in generations - replaces time_converter
    dg = dg.in_generations()

    # get classified discrete demographic events
    dg_events = dg.discrete_demographic_events()

    idmap = _build_deme_id_to_int_map(dg)
    initial_sizes = _get_initial_deme_sizes(dg, idmap)
    Nref = _get_ancestral_population_size(dg)

    burnin_generation = int(np.rint(burnin * Nref))
    model_times = _get_model_times(dg)

    events = _Fwdpy11Events(idmap=idmap)

    _set_initial_migration_matrix(dg, idmap, events)
    _process_all_epochs(dg, idmap, model_times, burnin_generation, events)
    _process_migrations(dg, idmap, model_times, burnin_generation, events)
    _process_pulses(dg, idmap, model_times, burnin_generation, events)
    _process_admixtures(dg, dg_events, idmap, model_times, burnin_generation, events)
    _process_mergers(dg, dg_events, idmap, model_times, burnin_generation, events)
    _process_splits(dg, dg_events, idmap, model_times, burnin_generation, events)
    _process_branches(dg, dg_events, idmap, model_times, burnin_generation, events)

    if dg.doi != "None":
        doi = dg.doi
    else:
        doi = None

    return DemographicModelDetails(
        model=events.build_model(),
        name=dg.description,
        source=source,
        parameters=None,
        citation=DemographicModelCitation(DOI=doi, full_citation=None, metadata=None),
        metadata={
            "deme_labels": {j: i for i, j in idmap.items()},
            "initial_sizes": initial_sizes,
            "burnin_time": burnin_generation,
            "total_simulation_length": burnin_generation + model_times.model_duration,
        },
    )


@attr.s(frozen=True, auto_attribs=True)
class _ModelTimes(object):
    """
    These are in units of the deme graph
    and increase backwards into the past.
    """

    model_start_time: demes.demes.Time
    model_end_time: demes.demes.Time
    model_duration: int = attr.ib(validator=attr.validators.instance_of(int))


@attr.s(auto_attribs=True)
class _MigrationRateChange(object):
    """
    Use to make registry of migration rate changes.

    """

    when: int = attr.ib(
        validator=[demes.demes.non_negative, attr.validators.instance_of(int)]
    )
    source: int = attr.ib(
        validator=[demes.demes.non_negative, attr.validators.instance_of(int)]
    )
    destination: int = attr.ib(
        validator=[demes.demes.non_negative, attr.validators.instance_of(int)]
    )
    rate_change: float = attr.ib(validator=[attr.validators.instance_of(float)])
    from_deme_graph: bool = attr.ib(validator=attr.validators.instance_of(bool))


@class_decorators.attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True)
class _Fwdpy11Events(object):
    """
    One stop shop for adding things we support in future versions.

    This class creates some redundancy with fwdpy11.DiscreteDemography,
    but that class is "frozen", so we need a mutable version.
    """

    mass_migrations: List[MassMigration] = attr.Factory(list)
    set_deme_sizes: List[SetDemeSize] = attr.Factory(list)
    set_growth_rates: List[SetExponentialGrowth] = attr.Factory(list)
    set_selfing_rates: List[SetSelfingRate] = attr.Factory(list)
    idmap: Dict = None

    # The initial continuous migration matrix
    initial_migmatrix: Optional[MigrationMatrix] = None
    # The migration matrix that we update to get changes in migration rates
    migmatrix: Optional[MigrationMatrix] = None

    # The following do not correspond to fwdpy11 event types.
    migration_rate_changes: List[_MigrationRateChange] = attr.Factory(list)
    # deme_extinctions: List[_DemeExtinctionEvent] = attr.Factory(list)

    def _update_changes_at_m(self, changes_at_m, migration_rate_change):
        # tally changes
        if migration_rate_change.from_deme_graph:
            changes_at_m[0][migration_rate_change.destination][
                migration_rate_change.source
            ] += migration_rate_change.rate_change
            if migration_rate_change.destination != migration_rate_change.source:
                changes_at_m[0][migration_rate_change.destination][
                    migration_rate_change.destination
                ] -= migration_rate_change.rate_change
        else:
            changes_at_m[1][migration_rate_change.destination][
                migration_rate_change.source
            ] += migration_rate_change.rate_change
            changes_at_m[1][migration_rate_change.destination][
                migration_rate_change.destination
            ] -= migration_rate_change.rate_change

    def _update_continuous_mass_migrations(self, changes_at_m, M_cont, M_mass):
        M_cont += changes_at_m[0]
        M_mass += changes_at_m[1]
        return M_cont, M_mass

    def _migration_matrix_from_partition(self, M_cont, M_mass):
        new_migmatrix = np.diag(np.diag(M_mass)).dot(M_cont) + (
            M_mass - np.diag(np.diag(M_mass))
        )
        return new_migmatrix

    def _build_migration_rate_changes(self) -> List[SetMigrationRates]:
        # We track the coninuous migration rates, and then augment with a matrix that
        # specifies changes to migration due to "instantaneous" events. The
        # instantaneous migration matrix is typically just the identity matrix,
        # (i.e. uses continuous rates, but off diag elements scale continuous rates
        # and add to ancestry source from the off diagonal source column)
        # but for some generations has ancestry pointing to different demes due
        # to pulse, split, etc events
        if self.migmatrix is not None:
            M_cont = copy.deepcopy(self.migmatrix)
        else:
            M_cont = np.eye(len(self.idmap))
        M_mass = np.eye(len(self.idmap))

        set_migration_rates: List[SetMigrationRates] = []

        self.migration_rate_changes = sorted(
            self.migration_rate_changes,
            key=lambda x: (x.when, x.destination, x.source, x.from_deme_graph),
        )
        # self.deme_extinctions = sorted(
        #    self.deme_extinctions, key=lambda x: (x.when, x.deme)
        # )
        m = 0
        changes_at_m = [
            np.zeros((len(self.idmap), len(self.idmap))),  # from Graph
            np.zeros((len(self.idmap), len(self.idmap))),  # not from Graph
        ]
        while m < len(self.migration_rate_changes):
            # gather all migration rate changes and extinction events
            # that occur at a given time
            self._update_changes_at_m(changes_at_m, self.migration_rate_changes[m])
            mm = m + 1

            while (
                mm < len(self.migration_rate_changes)
                and self.migration_rate_changes[mm].when
                == self.migration_rate_changes[m].when
            ):
                self._update_changes_at_m(changes_at_m, self.migration_rate_changes[mm])
                mm += 1

            # update M_cont and M_mass
            M_cont, M_mass = self._update_continuous_mass_migrations(
                changes_at_m, M_cont, M_mass
            )
            # get the new migration matrix
            new_migmatrix = self._migration_matrix_from_partition(M_cont, M_mass)
            # for any rows that don't match, add a fwdpy11.SetMigrationRate
            for i in range(len(self.idmap)):
                if np.any(self.migmatrix[i] != new_migmatrix[i]):
                    set_migration_rates.append(
                        SetMigrationRates(
                            self.migration_rate_changes[m].when, i, new_migmatrix[i]
                        )
                    )

            self.migmatrix = new_migmatrix

            m = mm
            # reset changes
            changes_at_m = [
                np.zeros((len(self.idmap), len(self.idmap))),  # from Graph
                np.zeros((len(self.idmap), len(self.idmap))),  # not from Graph
            ]

        return set_migration_rates

    ### now unused ??
    def build_model(self) -> DiscreteDemography:
        set_migration_rates = self._build_migration_rate_changes()
        return DiscreteDemography(
            mass_migrations=self.mass_migrations,
            set_deme_sizes=self.set_deme_sizes,
            set_growth_rates=self.set_growth_rates,
            set_selfing_rates=self.set_selfing_rates,
            migmatrix=self.initial_migmatrix,
            set_migration_rates=set_migration_rates,
        )


def _build_deme_id_to_int_map(dg: demes.Graph) -> Dict:
    """
    Convert the string input ID to output integer values.

    For sanity, the output values will be in increasing
    order of "deme origin" times.

    We rely on the epoch times being strictly sorted
    past-to-present in the YAML.  If there are ties,
    populations are sorted lexically by ID within
    start_time.
    """
    temp = []
    for deme in dg.demes:
        assert len(deme.epochs) > 0
        temp.append((deme.epochs[0].start_time, deme.name))

    temp = sorted(temp, key=lambda x: (-x[0], x[1]))

    return {j[1]: i for i, j in enumerate(temp)}


def _get_initial_deme_sizes(dg: demes.Graph, idmap: Dict) -> Dict:
    """
    Build a map of a deme's integer label to its size
    at the start of the simulation for all demes whose
    start_time equals inf.
    """
    otime = _get_most_ancient_deme_start_time(dg)
    rv = dict()
    for deme in dg.demes:
        if deme.epochs[0].start_time == otime:
            rv[idmap[deme.name]] = deme.epochs[0].start_size

    if len(rv) == 0:
        raise RuntimeError("could not determine initial deme sizes")

    return rv


def _get_most_ancient_deme_start_time(dg: demes.Graph) -> demes.demes.Time:
    return max([d.start_time for d in dg.demes])


def _get_most_recent_deme_end_time(dg: demes.Graph) -> demes.demes.Time:
    return min([d.end_time for d in dg.demes])


def _get_model_times(dg: demes.Graph) -> _ModelTimes:
    """
    In units of dg.time_units, obtain the following:

    1. The time when the demographic model starts.
    2. The time when it ends.
    3. The total simulation length.

    """
    # FIXME: this function isn't working well.
    # For example, twodemes.yml and twodemes_one_goes_away.yml
    # both break it.
    oldest_deme_time = _get_most_ancient_deme_start_time(dg)
    most_recent_deme_end = _get_most_recent_deme_end_time(dg)

    model_start_time = oldest_deme_time
    if oldest_deme_time == math.inf:
        # We want to find the time of first event or
        # the first demographic change, which is when
        # burnin will end. To do this, get a list of
        # first size change for all demes with inf
        # start time, and the start time for all other
        # demes, and take max of those.
        ends_inf = [d.epochs[0].end_time for d in dg.demes if d.start_time == math.inf]
        starts = [d.start_time for d in dg.demes if d.start_time != math.inf]
        mig_starts = [m.start_time for m in dg.migrations if m.start_time != math.inf]
        mig_ends = [m.end_time for m in dg.migrations if m.start_time == math.inf]
        pulse_times = [p.time for p in dg.pulses]
        model_start_time = max(ends_inf + starts + mig_starts + mig_ends + pulse_times)

    if most_recent_deme_end != 0:
        model_duration = model_start_time - most_recent_deme_end
    else:
        model_duration = model_start_time

    return _ModelTimes(
        model_start_time=model_start_time,
        model_end_time=most_recent_deme_end,
        model_duration=int(np.rint(model_duration)),
    )


def _get_ancestral_population_size(dg: demes.Graph) -> int:
    """
    Need this for the burnin time.

    If there are > 1 demes with the same most ancient start_time,
    then the ancestral size is considered to be the size
    of all those demes (size of ancestral metapopulation).
    """
    oldest_deme_time = _get_most_ancient_deme_start_time(dg)

    rv = sum(
        [
            e.start_size
            for d in dg.demes
            for e in d.epochs
            if e.start_time == oldest_deme_time
        ]
    )
    if rv == 0:
        raise RuntimeError("could not determinine ancestral metapopulation size")
    return rv


def _set_initial_migration_matrix(
    dg: demes.Graph, idmap: Dict, events: _Fwdpy11Events
) -> None:
    """
    Set any migration rates that have start time of inf. More
    recent migration rate changes or start time set the model
    start time at or before that time
    """
    if len(idmap) > 1:
        migmatrix = np.zeros((len(idmap), len(idmap)))
        for deme_id, ii in idmap.items():
            if dg[deme_id].start_time == math.inf:
                migmatrix[ii, ii] = 1.0
        if len(dg.migrations) > 0:
            for m in dg.migrations:
                if m.start_time == math.inf:
                    migmatrix[idmap[m.dest]][idmap[m.source]] += m.rate
                    migmatrix[idmap[m.dest]][idmap[m.dest]] -= m.rate

        events.migmatrix = migmatrix
        events.initial_migmatrix = migmatrix


##
## Process deme epochs, migrations, and discrete demographic events
##


def _process_epoch(
    deme_id: str,
    e: demes.Epoch,
    idmap: Dict,
    model_times: _ModelTimes,
    burnin_generation: int,
    events: _Fwdpy11Events,
) -> None:
    """
    Can change sizes, cloning rates, and selfing rates.

    Since fwdpy11 currently doesn't understand cloning, we need
    to raise an error if the rate is not None or nonzero.
    """
    if e.start_time != math.inf:
        when = burnin_generation + int(model_times.model_start_time - e.start_time)
    else:
        when = 0

    if e.selfing_rate is not None:
        events.set_selfing_rates.append(
            SetSelfingRate(when=when, deme=idmap[deme_id], S=e.selfing_rate)
        )

    if e.cloning_rate is not None:
        if e.cloning_rate > 0.0:
            raise ValueError("fwdpy11 does not currently support cloning rates > 0.")

    if e.start_time != math.inf:
        # Handle size change functions
        events.set_deme_sizes.append(
            SetDemeSize(
                when=when - 1,
                deme=idmap[deme_id],
                new_size=e.start_size,
            )
        )
        if e.end_size != e.start_size:
            if e.size_function != "exponential":
                raise ValueError(
                    f"Size change function must be exponential.  We got {e.size_function}"
                )
            G = exponential_growth_rate(
                e.start_size, e.end_size, int(np.rint(e.time_span))
            )
            events.set_growth_rates.append(
                SetExponentialGrowth(when=when, deme=idmap[deme_id], G=G)
            )


def _process_all_epochs(
    dg: demes.Graph,
    idmap: Dict,
    model_times: _ModelTimes,
    burnin_generation: int,
    events: _Fwdpy11Events,
):
    """
    Processes all epochs of all demes to set sizes and selfing rates.
    """
    for deme in dg.demes:
        for e in deme.epochs:
            _process_epoch(
                deme.name,
                e,
                idmap,
                model_times,
                burnin_generation,
                events,
            )
        # if a deme starts more recently than math.inf, we have to
        # turn on migration in that deme with a diagonal element to 1
        if deme.start_time < math.inf:
            events.migration_rate_changes.append(
                _MigrationRateChange(
                    when=burnin_generation
                    + int(model_times.model_start_time - deme.start_time),
                    source=idmap[deme.name],
                    destination=idmap[deme.name],
                    rate_change=1.0,
                    from_deme_graph=True,
                )
            )

        # if a deme ends before time zero, we set diag entry in migmatrix to 0
        if deme.end_time > 0:
            events.migration_rate_changes.append(
                _MigrationRateChange(
                    when=burnin_generation
                    + int(model_times.model_start_time - deme.end_time),
                    source=idmap[deme.name],
                    destination=idmap[deme.name],
                    rate_change=-1.0,
                    from_deme_graph=True,
                )
            )

        # if deme ends before time zero, we set set its size to zero
        # we proces deme extintions here instead of in the events
        if deme.end_time > 0:
            events.set_deme_sizes.append(
                SetDemeSize(
                    when=burnin_generation
                    + int(model_times.model_start_time - deme.end_time),
                    deme=idmap[deme.name],
                    new_size=0,
                )
            )


def _process_migrations(
    dg: demes.Graph,
    idmap: Dict,
    model_times: _ModelTimes,
    burnin_generation: int,
    events: _Fwdpy11Events,
) -> None:
    """
    Make a record of everything in dg.migrations

    When a migration rate has an end time > 0, it gets entered twice.
    """
    for m in dg.migrations:
        if m.start_time < math.inf:
            when = burnin_generation + int(model_times.model_start_time - m.start_time)
            try:
                events.migration_rate_changes.append(
                    _MigrationRateChange(
                        when=when,
                        source=idmap[m.source],
                        destination=idmap[m.dest],
                        rate_change=m.rate,
                        from_deme_graph=True,
                    )
                )
            except AttributeError:
                for source, dest in itertools.permutations(m.demes, 2):
                    events.migration_rate_changes.append(
                        _MigrationRateChange(
                            when=when,
                            source=idmap[source],
                            destination=idmap[dest],
                            rate_change=m.rate,
                            from_deme_graph=True,
                        )
                    )
        if m.end_time > 0:
            when = burnin_generation + int(model_times.model_start_time - m.end_time)
            try:
                events.migration_rate_changes.append(
                    _MigrationRateChange(
                        when=when,
                        source=idmap[m.source],
                        destination=idmap[m.dest],
                        rate_change=-m.rate,
                        from_deme_graph=True,
                        )
                    )
            except AttributeError:
                for source, dest in itertools.permutations(m.demes, 2):
                    events.migration_rate_changes.append(
                        _MigrationRateChange(
                            when=when,
                            source=idmap[source],
                            destination=idmap[dest],
                            rate_change=-m.rate,
                            from_deme_graph=True,
                        )
                    )


def _process_pulses(
    dg: demes.Graph,
    idmap: Dict,
    model_times: _ModelTimes,
    burnin_generation: int,
    events: _Fwdpy11Events,
) -> None:
    for p in dg.pulses:
        when = burnin_generation + int(model_times.model_start_time - p.time)
        events.migration_rate_changes.append(
            _MigrationRateChange(
                when=when - 1,
                source=idmap[p.source],
                destination=idmap[p.dest],
                rate_change=p.proportion,
                from_deme_graph=False,
            )
        )
        events.migration_rate_changes.append(
            _MigrationRateChange(
                when=when,
                source=idmap[p.source],
                destination=idmap[p.dest],
                rate_change=-p.proportion,
                from_deme_graph=False,
            )
        )


def _process_admixtures(
    dg: demes.Graph,
    dg_events: Dict,
    idmap: Dict,
    model_times: _ModelTimes,
    burnin_generation: int,
    events: _Fwdpy11Events,
) -> None:
    for a in dg_events["admixtures"]:
        when = burnin_generation + int(model_times.model_start_time - a.time)
        for parent, proportion in zip(a.parents, a.proportions):
            events.migration_rate_changes.append(
                _MigrationRateChange(
                    when=when - 1,
                    source=idmap[parent],
                    destination=idmap[a.child],
                    rate_change=proportion,
                    from_deme_graph=False,
                )
            )
            events.migration_rate_changes.append(
                _MigrationRateChange(
                    when=when,
                    source=idmap[parent],
                    destination=idmap[a.child],
                    rate_change=-proportion,
                    from_deme_graph=False,
                )
            )


def _process_mergers(
    dg: demes.Graph,
    dg_events: Dict,
    idmap: Dict,
    model_times: _ModelTimes,
    burnin_generation: int,
    events: _Fwdpy11Events,
) -> None:
    for m in dg_events["mergers"]:
        when = burnin_generation + int(model_times.model_start_time - m.time)
        for parent, proportion in zip(m.parents, m.proportions):
            events.migration_rate_changes.append(
                _MigrationRateChange(
                    when=when - 1,
                    source=idmap[parent],
                    destination=idmap[m.child],
                    rate_change=proportion,
                    from_deme_graph=False,
                )
            )
            events.migration_rate_changes.append(
                _MigrationRateChange(
                    when=when,
                    source=idmap[parent],
                    destination=idmap[m.child],
                    rate_change=-proportion,
                    from_deme_graph=False,
                )
            )


def _process_splits(
    dg: demes.Graph,
    dg_events: Dict,
    idmap: Dict,
    model_times: _ModelTimes,
    burnin_generation: int,
    events: _Fwdpy11Events,
) -> None:
    """
    A split is a "sudden" creation of > 1 offspring deme
    from a parent and the parent ceases to exist.

    Given that there is no proportions attribute, we infer/assume
    (danger!) that each offspring deme gets 100% of its ancestry
    from the parent.
    """
    for s in dg_events["splits"]:
        when = burnin_generation + int(model_times.model_start_time - s.time)
        for c in s.children:
            # one generation of migration to move lineages from parent to children
            events.migration_rate_changes.append(
                _MigrationRateChange(
                    when=when - 1,
                    source=idmap[s.parent],
                    destination=idmap[c],
                    rate_change=1.0,
                    from_deme_graph=False,
                )
            )
            # turn off that migration after one generation
            events.migration_rate_changes.append(
                _MigrationRateChange(
                    when=when,
                    source=idmap[s.parent],
                    destination=idmap[c],
                    rate_change=-1.0,
                    from_deme_graph=False,
                )
            )


def _process_branches(
    dg: demes.Graph,
    dg_events: Dict,
    idmap: Dict,
    model_times: _ModelTimes,
    burnin_generation: int,
    events: _Fwdpy11Events,
) -> None:
    """
    A branch creates a child deme with 100% ancestry from the parent.
    The parent continues to exist.

    The 1-to-1 relationship between parent and child means 100% of the
    child's ancestry is from parent.
    """
    for b in dg_events["branches"]:
        when = burnin_generation + int(model_times.model_start_time - b.time)
        # turn on migration for one generation at "when"
        events.migration_rate_changes.append(
            _MigrationRateChange(
                when=when - 1,
                source=idmap[b.parent],
                destination=idmap[b.child],
                rate_change=1.0,
                from_deme_graph=False,
            )
        )
        # end that migration after one generation
        events.migration_rate_changes.append(
            _MigrationRateChange(
                when=when,
                source=idmap[b.parent],
                destination=idmap[b.child],
                rate_change=-1.0,
                from_deme_graph=False,
            )
        )
